/******************************************************************************
Copyright (C) 2002 Red Hat, Inc.

This file is part of op_visualise, a part of Oprofile.

This file may be redistributed and/or modifed under the terms of the GNU
General Public License as published by the Free Software Foundation; either
version 2, or (at your option) any later version.

op_visualise is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details. 

op_visualise was written, and is maintained, by Chris Moller,
Red Hat Inc, cmoller@redhat.com

This file provides functions that initialise op_visualise.

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define _GNU_SOURCE	/* enables GNU C extensions */
#include <getopt.h>
#define __USE_ISOC99 1
#include <math.h>

#define DO_INITIALISE
#include "op_visualise.h"

#include "function-icon.h"
#include "file-icon.h"
#include "system-icon.h"

static GtkWidget * toolbar_function_window;
static GtkWidget * toolbar_system_window;
static GtkWidget * toolbar_file_window;
static GtkWidget * toolbar;

static gchar * info_titles[] = {
  "Samples",
  "Executable",
#ifdef USE_SOURCES_BUTTON
  "Sources"
#endif
};

typedef enum {
  TOOLBAR_NULL,
  TOOLBAR_FUNCTIONS,
  TOOLBAR_SYSTEM,
  TOOLBAR_FILE
} toolbar_e;

double position_d;
static void toolbar_callback(GtkWidget * button __attribute__ ((unused)), gpointer data);

void
update_labels()
{
  if (left_label_text)  free(left_label_text);
  if (right_label_text) free(right_label_text);
  asprintf(&left_label_text,  "| %08lx", lrint(min_ruler_d));
  asprintf(&right_label_text, "%08lx |", lrint(max_ruler_d));
  gtk_label_set(GTK_LABEL(left_label),  left_label_text);
  gtk_label_set(GTK_LABEL(right_label), right_label_text);
}

#define SHORT_SAMPLES_DIRECTORY		(int)'S'
#define LONG_SAMPLES_DIRECTORY		"samples-directory"
#define SHORT_SAMPLES_FILE		(int)'s'
#define LONG_SAMPLES_FILE		"samples-file"
#define SHORT_BINARIES_DIRECTORY	(int)'B'
#define LONG_BINARIES_DIRECTORY		"binaries-directory"
#define SHORT_EXECUTABLE		(int)'b'
#define LONG_EXECUTABLE			"executable"
#define SHORT_SOURCE_DIRECTORY		(int)'d'
#define LONG_SOURCE_DIRECTORY		"source-directory"
#define SHORT_HELP			(int)'h'
#define LONG_HELP			"help"
#define SHORT_INITS_FILE		(int)'i'
#define LONG_INITS_FILE			"initialisation-file"

static char usage_text[] = {
#include "op_visualise_help.h"
};

static void
show_usage()
{
  GtkWidget     * view	 = gtk_text_view_new();
  GtkTextBuffer * buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
  GtkWidget     * window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
  GtkWidget     * scroll = gtk_scrolled_window_new (NULL, NULL);
  GtkWidget     * vbox    = gtk_vbox_new(FALSE, 5);

  gtk_text_buffer_set_text(buffer, usage_text, -1);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(view), FALSE);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(view), TRUE);
  gtk_widget_show(view);
  
  gtk_widget_set_uposition(window, 150, 450 );
  gtk_widget_set_usize (window, 500, 250);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_signal_connect(GTK_OBJECT(window),
		     "destroy",
	             GTK_SIGNAL_FUNC(gtk_widget_destroy),
		     NULL);
  gtk_container_add(GTK_CONTAINER(window), vbox);
  gtk_widget_show(vbox);

  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll),
		                  GTK_POLICY_AUTOMATIC,
				  GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scroll), view);
  gtk_widget_show(scroll);
  gtk_box_pack_start(GTK_BOX(vbox), scroll, TRUE, TRUE, 0);
  gtk_widget_show(window);
}

static void
show_about()
{
  GtkWidget *  dialog = gtk_message_dialog_new (GTK_WINDOW(main_window),
		                                GTK_DIALOG_DESTROY_WITH_PARENT,
			                        GTK_MESSAGE_INFO,
	                                        GTK_BUTTONS_CLOSE,
	                                        "\
op_visualise: an Oprofile data visualisation tool.\n\
\n\
Copyright (c) 2002 by Red Hat, Inc\n\
\n\
This software may be freely redistributed under the terms\n\
of the GNU General Public License.\n\
\n\
Written by Chris Moller (cmoller@redhat.com), Red Hat, Inc.");

  gtk_dialog_run (GTK_DIALOG (dialog));
  gtk_widget_destroy (dialog);
}

static int
symbol_list_search_fcn(const void * a, const void * b)
{
  symvalue bb = (*(symbol_info **)b)->value;
  return ((bfd_vma)a < bb) ? -1 : (((bfd_vma)a > bb) ? 1 : 0);
}

#define POSITIONS_HISTORY_SLOTS 5
static gdouble * positions_history = NULL;
static int positions_history_count = 0;
static gdouble positions_history_sum = 0;
static gboolean previous_position_valid = FALSE;



static void
show_stats(gdouble position)
{
  int i;
  static gdouble previous_position;
  bfd_vma d_lpos = 0;
  bfd_vma lpos;
  bfd_vma spos;
  const char *filename, *functionname;
  unsigned int lineno;
  char * posn;
  char * pposn;
  int lookup_hit;
  int do_redraw;
  static int old_lookup_hit = -1;
  
  if (positions_history) {
    if (previous_position_valid) {
      if (positions_history_count >= POSITIONS_HISTORY_SLOTS) 
	positions_history_sum -= positions_history[POSITIONS_HISTORY_SLOTS - 1];
      memmove(&positions_history[1], &positions_history[0],
	      positions_history_count * sizeof(gdouble));
      positions_history[0] = abs(position - previous_position);
      positions_history_sum += positions_history[0];
      if (positions_history_count < POSITIONS_HISTORY_SLOTS) positions_history_count++;
      d_lpos = lrint(positions_history_sum/(2.0 * (gdouble)positions_history_count));
    }
    previous_position = position;
    previous_position_valid = TRUE;
  }
    
  lpos = (bfd_vma)lrint((double)position);

  do_redraw = 0;
  if (-1 != old_lookup_hit) {
    display_callback(op_samples[old_lookup_hit].loc,
		     op_samples[old_lookup_hit].value,
		     drawing_area->style->white_gc);
    do_redraw = 1;
  }

  posn = NULL;
  pposn = NULL;
  if (ibfd && stree_valid) {
    if (-1 != (lookup_hit = samples_locate_nearest_in_bound(lpos, d_lpos))) {
      display_callback(op_samples[lookup_hit].loc,
		       op_samples[lookup_hit].value,
		       green_gc);
      old_lookup_hit = lookup_hit;
      do_redraw = 1;
    }

    for ( i = 0; i < text_section_next; i++) {
      if ((lpos >= (spos = bfd_get_section_vma(ibfd, text_section_list[i]))) &&
	  (lpos < (spos + bfd_section_size(ibfd, text_section_list[i])))) {
	if (bfd_find_nearest_line(ibfd, text_section_list[i], syms,
				  lpos - bfd_get_section_vma(ibfd, text_section_list[i]),
				  &filename, &functionname, &lineno)) {
	  if (lookup_hit != -1)
	    asprintf(&posn, "%#010lx [%u]: %s -- %s:%d (%s)",
		     op_samples[lookup_hit].loc, op_samples[lookup_hit].value,
		     bfd_get_section_name(ibfd, text_section_list[i]), 
		     filename, lineno, functionname);
	  else
	    asprintf(&posn, "%#010lx: %s -- %s:%d (%s)",
		     lpos, bfd_get_section_name(ibfd, text_section_list[i]),
		     filename, lineno, functionname);
	}
	else {
	  if (lookup_hit != -1)
	    asprintf(&posn, "%#010lx [%u]: %s",
		     op_samples[lookup_hit].loc, op_samples[lookup_hit].value,
		     bfd_get_section_name(ibfd, text_section_list[i]));
	  else
	    asprintf(&posn, "%#010lx: %s",
		     lpos, bfd_get_section_name(ibfd, text_section_list[i]));
	}
	break;
      }
    }
  }
  if (do_redraw) gtk_widget_draw (drawing_area, NULL);

  if (symbol_list) {
    symbol_info ** symbol_entry_p;

    if ((symbol_entry_p = bsearch((const void *)lpos, symbol_list,
				 sym_count, sizeof(symbol_info *), symbol_list_search_fcn))) {
      asprintf(&pposn,"%s [symbol %s]", posn ? : "", (*symbol_entry_p)->name);
    }
  }

  if (pposn) gtk_label_set(GTK_LABEL(status_bar), pposn);
  else if (posn) gtk_label_set(GTK_LABEL(status_bar), posn);
  if (posn) free(posn);
  if (pposn) free(pposn);
}

static gint
motion_notify_event( GtkWidget * widget, GdkEventMotion * event)
{
  int x,y;
  GdkModifierType state;
  gboolean rc = TRUE;
  gdouble position;

  if (long_op_in_progress) return rc;
  if (event->is_hint) gdk_window_get_pointer (event->window, &x, &y, &state);
  else {
    x = event->x;
    y = event->y;
    state = event->state;
  }

  position = (gdouble)(x)/(gdouble)(widget->allocation.width);
  position *= (max_ruler_d - min_ruler_d);
  position += min_ruler_d;
  show_stats(position);
  
  return rc;
}

static void
set_main_title(char * str)
{
  static char * title_str = 0;

  if (title_str) free(title_str);
  asprintf(&title_str,"Oprofile %s", str);
  gtk_window_set_title(GTK_WINDOW(main_window), title_str);
}

static GtkItemFactoryEntry menu_items[] = {
  {"/_File",			"<control>F",	NULL,			0,		"<Branch>",	NULL},
  {"/File/Set Sample Dir...",	"<control>D",	handle_select_file,	SAMPLES_DIR,	NULL, 		NULL},
  {"/File/Set Executables _Root...",	"<control>R",	handle_select_file,	BINARIES_DIR,	NULL,	NULL},
#if 0
  {"/File/Select Sources...",	"<control>D",	handle_select_file,	SOURCES_DIR,	NULL,		NULL},
#endif
  {"/File/sep0",		NULL,		NULL,			0,		"<Separator>", 	NULL},
  {"/File/Open _Sample...",	"<control>S",	handle_open_sample,	SAMPLES_FILE,	NULL,		NULL},
  {"/File/Select _Binary...",	"<control>B",	handle_open_sample,	BINARY_FILE,	NULL,		NULL},
  {"/File/sep1",		NULL,		NULL,			0,		"<Separator>", 	NULL},
#ifdef ENABLE_PLOT
  {"/File/_Print Chart...",	"<control>P",	handle_print_file,	PRINT_FILE,	NULL,		NULL},
  {"/File/sep2",		NULL,		NULL,			0,		"<Separator>", 	NULL},
#endif
  {"/File/Quit",		"<control>Q",	gtk_main_quit,		0,		NULL,		NULL},
  {"/_Help",			"<control>H",	NULL,			0,		"<Branch>",	NULL},
  {"/_Help/Usage",		"NULL",		show_usage,		0,		NULL,		NULL},
  {"/Help/sep1",		NULL,		NULL,			0,		"<Separator>", 	NULL},
  {"/_Help/About",		"NULL",		show_about,		0,		NULL,		NULL},
};

static GtkItemFactory * item_factory;

static void
get_main_menu(GtkWidget * window, GtkWidget ** menu_bar)
{
  GtkAccelGroup	* accel_group;
  gint nmenu_items = sizeof (menu_items) / sizeof(GtkItemFactoryEntry);

  accel_group = gtk_accel_group_new();
  item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", accel_group);
  gtk_item_factory_create_items(item_factory, nmenu_items, menu_items, NULL);
  gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);
  if (menu_bar) *menu_bar = gtk_item_factory_get_widget (item_factory, "<main>");
}

static void
info_clicked(GtkWidget	* table __attribute__ ((unused)), gpointer pdata)
{
  switch((long int)pdata) {
  case SAMPLES_FILE:
    handle_open_sample(NULL, SAMPLES_FILE, NULL);
    break;
  case BINARY_FILE:
    handle_open_sample(NULL, BINARY_FILE, NULL);
    break;
#ifdef USE_SOURCES_BUTTON
  case SOURCES_DIR:
    handle_select_file(NULL, SOURCES_DIR, NULL);
    break;
#endif
  }
}

static GdkColor red    = {0, 65535,     0, 0};
static GdkColor green  = {0,     0, 65535, 0};
static GdkColor yellow = {0, 65535, 65535, 0};

static gint
configure_event( GtkWidget * widget, GdkEventConfigure * event __attribute__ ((unused)))
{
  if (pixmap) gdk_pixmap_unref(pixmap);

  pixmap = gdk_pixmap_new(widget->window,
                          widget->allocation.width,
                          widget->allocation.height,
                          -1);

  red_gc = gdk_gc_new(widget->window);
  gdk_gc_set_rgb_fg_color(red_gc, &red);

  green_gc = gdk_gc_new(widget->window);
  gdk_gc_set_rgb_fg_color(green_gc, &green);

  yellow_gc = gdk_gc_new(widget->window);
  gdk_gc_set_rgb_fg_color(yellow_gc, &yellow);

  gdk_draw_rectangle (pixmap,
                      widget->style->black_gc,
                      TRUE,
                      0, 0,
                      widget->allocation.width,
                      widget->allocation.height);

  gtk_widget_draw (widget, NULL);
  maybe_visualise_data();

  return TRUE;
}

static gint
expose_event( GtkWidget * widget, GdkEventExpose * event)
{
  gdk_draw_pixmap(widget->window,
                  widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                  pixmap,
                  event->area.x, event->area.y,
                  event->area.x, event->area.y,
                  event->area.width, event->area.height);

  return FALSE;
}

typedef struct {
  char * key;
  int (*action)();
  int  code;
} init_procs_s;

static int
init_proc(int code, char * optarg)
{
  int rc = 1;
  
  switch(code) {
  case SHORT_BINARIES_DIRECTORY:
    files_in_use[BINARIES_DIR] = strdup(optarg);
    break;
  case SHORT_SAMPLES_DIRECTORY:
    files_in_use[SAMPLES_DIR] = strdup(optarg);
    break;
  case SHORT_SAMPLES_FILE:
    files_in_use[SAMPLES_FILE] = strdup(optarg);
    file_changed |= (1 << SAMPLES_FILE);
    break;
  case SHORT_EXECUTABLE:
    files_in_use[BINARY_FILE] = strdup(optarg);
    file_changed |= (1 << BINARY_FILE);
    break;
  case SHORT_SOURCE_DIRECTORY:
    files_in_use[SOURCES_DIR] = strdup(optarg);
    break;
  default:
    rc = 0;
  }

  return rc;
}

init_procs_s init_procs[] = {
  {LONG_SAMPLES_DIRECTORY,	init_proc, SHORT_SAMPLES_DIRECTORY},
  {LONG_SAMPLES_FILE,		init_proc, SHORT_SAMPLES_FILE},
  {LONG_BINARIES_DIRECTORY,	init_proc, SHORT_BINARIES_DIRECTORY},
  {LONG_EXECUTABLE,		init_proc, SHORT_EXECUTABLE},
  {LONG_SOURCE_DIRECTORY,	init_proc, SHORT_SOURCE_DIRECTORY}   
};

static void 
parse_inits_file(char * inits_file)
{
  FILE * inits;
  char * key;
  char * value;
  unsigned int iv;

  if ((inits = fopen(inits_file, "r"))) {
    while (2 == (iv = fscanf(inits, " %a[-a-zA-Z] : %a[/.0-9a-zA-Z] \n", &key, &value))) {
      for (iv = 0; iv < sizeof(init_procs)/sizeof(init_procs_s); iv++) {
	if (!strcmp(key, init_procs[iv].key)) {
	  (*init_procs[iv].action)(init_procs[iv].code, value);
	  break;
	}
      }
      if (key)   free(key);
      if (value) free(value);
    }
    fclose(inits);
  }
}

static void
usage()
{
  fprintf(stderr,"Usage:\n\n\top_visualise [options]\n\nOptions:\n\n");
  
  fprintf(stderr,"\t-%c, --%s\t<file>\t\tSpecify samples directory\n",
	  SHORT_SAMPLES_DIRECTORY,  LONG_SAMPLES_DIRECTORY);
  fprintf(stderr,"\t\t(default %s\n",DEFAULT_SAMPLES_DIR);
  
  fprintf(stderr,"\t-%c, --%s\t<file>\t\tSpecify samples file\n",
	  SHORT_SAMPLES_FILE, LONG_SAMPLES_FILE);
  
  fprintf(stderr,"\t-%c, --%s\t<file>\t\tSpecify executables root directory\n",
	  SHORT_BINARIES_DIRECTORY,  LONG_BINARIES_DIRECTORY);
  fprintf(stderr,"\t\t(default %s\n",DEFAULT_BINARIES_DIR);
  
  fprintf(stderr,"\t-%c, --%s\t<file>\t\tSpecify executable file\n",
	  SHORT_EXECUTABLE, LONG_EXECUTABLE);
  
  fprintf(stderr,"\t-%c, --%s\t<file>\t\tSpecify source directory\n",
	  SHORT_SOURCE_DIRECTORY,  LONG_SOURCE_DIRECTORY);
  
  fprintf(stderr,"\t-%c, --%s\t\t<file>\t\tSpecify initialisation file.\n",
	  SHORT_INITS_FILE, LONG_INITS_FILE);
  fprintf(stderr,"\t\t(default %s\n",DEFAULT_INITS_FILE);
  
  fprintf(stderr,"\t-%c, --%s\t\t<file>\t\tPrint this usage.\n",
	  SHORT_HELP, LONG_HELP);
}

static gint
button_press_event( GtkWidget * widget __attribute__ ((unused)), GdkEventButton * event)
{
  double location = event->x/(gdouble)(widget->allocation.width);
  double position = min_ruler_d + location * (max_ruler_d - min_ruler_d);

  switch(event->button) {
  case 1:
    GTK_ADJUSTMENT(adjustment)->page_size /= 2.0;
    if (GTK_ADJUSTMENT(adjustment)->page_size < (gdouble)drawing_area->allocation.width)
      GTK_ADJUSTMENT(adjustment)->page_size = (gdouble)drawing_area->allocation.width;
    break;
  case 2:
    GTK_ADJUSTMENT(adjustment)->page_size = max_key_d - min_key_d;
    break;
  case 3:
    GTK_ADJUSTMENT(adjustment)->page_size *= 2.0;
    if (GTK_ADJUSTMENT(adjustment)->page_size > (max_key_d - min_key_d))
      GTK_ADJUSTMENT(adjustment)->page_size = max_key_d - min_key_d;
    break;
  }
  GTK_ADJUSTMENT(adjustment)->value = position - GTK_ADJUSTMENT(adjustment)->page_size * location;
  if (GTK_ADJUSTMENT(adjustment)->value < min_key_d) GTK_ADJUSTMENT(adjustment)->value = min_key_d;

  min_ruler_d = GTK_ADJUSTMENT(adjustment)->value;
  max_ruler_d = min_key_d + GTK_ADJUSTMENT(adjustment)->page_size;
  update_labels();
  position_d = position;
  gtk_adjustment_changed(GTK_ADJUSTMENT(adjustment));
  
  
  x_scale = ((double)drawing_area->allocation.width)/(max_ruler_d - min_ruler_d);
  
  draw_chart();

  return TRUE;
}

static void
adj_changed_event(GtkAdjustment * adjustment, gpointer user_data __attribute__ ((unused)))
{
  long int v_l;
  char * posn;

  min_ruler_d = adjustment->value;
  max_ruler_d = adjustment->value + adjustment->page_size;
  update_labels();
  position_d  = (max_ruler_d + min_ruler_d)/2.0;
  
  v_l = (min_ruler_d + max_ruler_d)/2.0;

  asprintf(&posn, "%#10lx", v_l);
  gtk_label_set(GTK_LABEL(status_bar), posn);
  free(posn);

  draw_chart();
}

int tool_wins_in_use = 0;

static char * toolbar_fn = 0;

#define TOOL_FUNCTION_WIN_ACTIVE (tool_wins_in_use & (1 << TOOLBAR_FUNCTIONS))
#define TOOL_SYSTEM_WIN_ACTIVE   (tool_wins_in_use & (1 << TOOLBAR_SYSTEM))
#define TOOL_FILE_WIN_ACTIVE     (tool_wins_in_use & (1 << TOOLBAR_FILE))

#define TOOL_FUNCTION_WIN_ACTIVATE (tool_wins_in_use |= (1 << TOOLBAR_FUNCTIONS))
#define TOOL_SYSTEM_WIN_ACTIVATE   (tool_wins_in_use |= (1 << TOOLBAR_SYSTEM))
#define TOOL_FILE_WIN_ACTIVATE     (tool_wins_in_use |= (1 << TOOLBAR_FILE))

#define TOOL_FUNCTION_WIN_DEACTIVATE (tool_wins_in_use &= ~(1 << TOOLBAR_FUNCTIONS))
#define TOOL_SYSTEM_WIN_DEACTIVATE   (tool_wins_in_use &= ~(1 << TOOLBAR_SYSTEM))
#define TOOL_FILE_WIN_DEACTIVATE     (tool_wins_in_use &= ~(1 << TOOLBAR_FILE))

static gboolean
composite_tree_extract(gpointer key __attribute__ ((unused)), 
		      gpointer value, 
		      gpointer model)
{
  GtkTreeIter iter;
  composite_s * fdp = (composite_s *)value;

  gtk_list_store_append (GTK_LIST_STORE (model), &iter);
  gtk_list_store_set (GTK_LIST_STORE (model), &iter, 
		  0, fdp->sample_fn, 
		  1, fdp->total_values,
		  2, 100.0 * ((double)(fdp->total_values))/((double)max_v),
		  -1);
  return FALSE;
}

static gboolean
function_tree_extract(gpointer key __attribute__ ((unused)), 
		      gpointer value, 
		      gpointer model)
{
  GtkTreeIter iter;
  function_data_s * fdp = (function_data_s *)value;

  gtk_list_store_append (GTK_LIST_STORE (model), &iter);
  gtk_list_store_set (GTK_LIST_STORE (model), &iter, 
		  0, fdp->function_name, 
		  1, fdp->total_hits,
		  2, 100.0 * ((double)(fdp->total_hits))/((double)total_hits),
		  -1);
  return FALSE;
}

static void
toolbar_sel_changed(GtkTreeSelection * selection,
                    gpointer data __attribute__ ((unused)))
{
  GtkTreeIter iter;
  GtkTreeModel * model;

  if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
    gtk_tree_model_get(model, &iter, 0, &toolbar_fn, -1);
    {
      function_data_s * hit;

      if (toolbar_fn) {
        switch ((toolbar_e)data) {
        case TOOLBAR_FUNCTIONS:
          if ((hit = g_tree_lookup(function_tree, toolbar_fn))) {
	    highlight_active = TRUE;
	    highlight_min = hit->min_loc;
	    highlight_max = hit->max_loc;
	    draw_chart();
          }
          break;
        case TOOLBAR_SYSTEM:
          break;
        case TOOLBAR_FILE:
          if ((hit = g_tree_lookup(file_tree, toolbar_fn))) {
	    highlight_active = TRUE;
	    highlight_min = hit->min_loc;
	    highlight_max = hit->max_loc;
	    draw_chart();
          }
          break;
        case TOOLBAR_NULL:
          break;
        }
      }
    }
  }
} 

static void
toolbar_destroy(GtkObject * object __attribute__ ((unused)),
	       gint        data)
{
  if (toolbar_fn) free(toolbar_fn);
  toolbar_fn = NULL;
  highlight_active = FALSE;
  switch ((toolbar_e)data) {
  case TOOLBAR_FUNCTIONS:
    TOOL_FUNCTION_WIN_DEACTIVATE;
    break;
  case TOOLBAR_SYSTEM:
    TOOL_SYSTEM_WIN_DEACTIVATE;
    break;
  case TOOLBAR_FILE:
    TOOL_FILE_WIN_DEACTIVATE;
    break;
  case TOOLBAR_NULL:
    break;
  }
  draw_chart();
}

static void
toolbar_close(GtkObject * object __attribute__ ((unused)),
	      gpointer data)
{
  gtk_widget_destroy(GTK_WIDGET(data));
}

static void
toolbar_callback(GtkWidget * button __attribute__ ((unused)), gpointer data)
{
  GtkWidget         * scroll;
  GtkWidget         * vbox;
  GtkWidget         * tree;
  GtkListStore      * model;
  GtkCellRenderer   * cell;
  GtkTreeViewColumn * column;
  GtkTreeSelection  * select;
  GtkWidget 	    * close;
  GtkWidget        ** toolbar_window_p = NULL;

  switch ((toolbar_e)data) {
  case TOOLBAR_FUNCTIONS:
    if (!function_tree  || TOOL_FUNCTION_WIN_ACTIVE) return;
    toolbar_window_p = &toolbar_function_window;
    break;
  case TOOLBAR_SYSTEM:
    if (!composite_tree || TOOL_SYSTEM_WIN_ACTIVE) return;
    toolbar_window_p = &toolbar_system_window;
    break;
  case TOOLBAR_FILE:
    if (!file_tree      || TOOL_FILE_WIN_ACTIVE) return;
    toolbar_window_p = &toolbar_file_window;
    break;
  case TOOLBAR_NULL:
    return;
    break;
  }

  *toolbar_window_p = gtk_window_new( GTK_WINDOW_TOPLEVEL);
  scroll  = gtk_scrolled_window_new (NULL, NULL);
  vbox    = gtk_vbox_new(FALSE, 5);
  tree    = gtk_tree_view_new ();
  model   = gtk_list_store_new (3, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_DOUBLE);
  close   = gtk_button_new_from_stock(GTK_STOCK_CLOSE);

  gtk_widget_set_uposition(*toolbar_window_p, 150, 450 );
  gtk_widget_set_usize (*toolbar_window_p, 500, 250);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_signal_connect(GTK_OBJECT(*toolbar_window_p),
		     "destroy",
	             GTK_SIGNAL_FUNC(toolbar_destroy),
		     data);
  gtk_container_add(GTK_CONTAINER(*toolbar_window_p), vbox);
  gtk_widget_show(vbox);

  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll),
		                  GTK_POLICY_AUTOMATIC,
				  GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scroll), tree);
  gtk_widget_show(tree);

  g_signal_connect(GTK_OBJECT(close), "clicked",
		   GTK_SIGNAL_FUNC (toolbar_close),
		   *toolbar_window_p);

  switch ((toolbar_e)data) {
  case TOOLBAR_FUNCTIONS:
    TOOL_FUNCTION_WIN_ACTIVATE;
    gtk_window_set_title(GTK_WINDOW(toolbar_function_window), "Functions");
    g_tree_foreach(function_tree, function_tree_extract, model);
    break;
  case TOOLBAR_SYSTEM:
    TOOL_SYSTEM_WIN_ACTIVATE;
    gtk_window_set_title(GTK_WINDOW(toolbar_system_window), "System");
    g_tree_foreach(composite_tree, composite_tree_extract, model);
    break;
  case TOOLBAR_FILE:
    TOOL_FILE_WIN_ACTIVATE;
    gtk_window_set_title(GTK_WINDOW(toolbar_file_window), "Files");
    g_tree_foreach(file_tree, function_tree_extract, model);
    break;
  case TOOLBAR_NULL:
    break;
  }
  gtk_tree_view_set_model(GTK_TREE_VIEW(tree), GTK_TREE_MODEL(model));
  g_object_unref(G_OBJECT(model));
  select = gtk_tree_view_get_selection(GTK_TREE_VIEW (tree));
  gtk_tree_selection_set_mode(select, GTK_SELECTION_SINGLE);
  g_signal_connect(G_OBJECT(select), "changed",
                   G_CALLBACK(toolbar_sel_changed),
                   data);
  
  cell    = gtk_cell_renderer_text_new ();
  column = gtk_tree_view_column_new_with_attributes ("Name",    cell, "text", 0, NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree),
	                      GTK_TREE_VIEW_COLUMN(column));

  cell    = gtk_cell_renderer_text_new ();
  column = gtk_tree_view_column_new_with_attributes ("Hits",    cell, "text", 1, NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree),
	                      GTK_TREE_VIEW_COLUMN(column));

  cell    = gtk_cell_renderer_text_new ();
  column = gtk_tree_view_column_new_with_attributes ("Percent", cell, "text", 2, NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree),
	                      GTK_TREE_VIEW_COLUMN(column));
  
  gtk_box_pack_start(GTK_BOX(vbox), scroll, TRUE, TRUE, 0);
  gtk_box_pack_end (GTK_BOX (vbox), close, FALSE, FALSE, 0);
  gtk_widget_show (close);
  gtk_widget_show(scroll);
  gtk_widget_show(*toolbar_window_p);
}

gint
main(int argc, gchar ** argv)
{                                  
  GtkWidget * main_vbox;
  char * inits_file = 0;
  
  static struct option long_options[] = {
    {LONG_BINARIES_DIRECTORY,	required_argument,	NULL,	SHORT_SAMPLES_DIRECTORY},
    {LONG_SAMPLES_DIRECTORY,	required_argument,	NULL,	SHORT_SAMPLES_DIRECTORY},
    {LONG_SAMPLES_FILE,		required_argument,	NULL,	SHORT_SAMPLES_FILE},
    {LONG_EXECUTABLE,		required_argument,	NULL,	SHORT_EXECUTABLE},
    {LONG_SOURCE_DIRECTORY,	required_argument,	NULL,	SHORT_SOURCE_DIRECTORY},
    {LONG_INITS_FILE,		required_argument,	NULL,	SHORT_INITS_FILE},
    {LONG_HELP,			no_argument,		NULL,	SHORT_HELP},
  };

  gtk_init(&argc, &argv);

#ifdef USE_ANNOTATION
  font = gdk_font_load("fixed");
#endif

  {
    int c;
    int option_index = 0;
    
    while (-1 != (c = getopt_long (argc, argv, "i:s:S:B:b:d:", long_options, &option_index))) {
      if (!init_proc(c, optarg)) {
	switch(c) {
	case SHORT_HELP:
	case '?':
	  usage();
	  exit(0);
	  break;
	case SHORT_INITS_FILE:
	  inits_file = strdup(optarg);
	  break;
	}
      }
    }
  }

  if (!files_in_use[SAMPLES_DIR]) files_in_use[SAMPLES_DIR] = strdup(DEFAULT_SAMPLES_DIR);
  if (!files_in_use[BINARIES_DIR]) files_in_use[BINARIES_DIR] = strdup(DEFAULT_BINARIES_DIR);
  if (!inits_file) inits_file = strdup(DEFAULT_INITS_FILE);

  parse_inits_file(inits_file);

  /* main window */
  
  main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect(GTK_OBJECT(main_window),
		     "destroy",
		     GTK_SIGNAL_FUNC(gtk_main_quit),
		     NULL);
  set_main_title("");
  gtk_widget_set_usize (GTK_WIDGET(main_window), main_window_width, main_window_height);


  /* main vbox */

  main_vbox = gtk_vbox_new(FALSE, 1);
  gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 5);
  gtk_container_add(GTK_CONTAINER(main_window), main_vbox);
  gtk_widget_show(main_vbox);


  /* main menu */

  {
    GtkWidget	 	* menu_bar;
    get_main_menu(main_window, &menu_bar);
    gtk_box_pack_start(GTK_BOX (main_vbox), menu_bar, FALSE, TRUE, 0);
    gtk_widget_show(menu_bar);
  }

  /* info bar */

  {
    GtkWidget * info_frame;
    GtkWidget * info_table;
    GtkWidget * button;
    long int i;
    long int ll = sizeof(info_titles)/sizeof(gchar *);
    
    info_frame = gtk_frame_new(NULL);
    gtk_container_set_border_width(GTK_CONTAINER(info_frame), 5);
    gtk_box_pack_start(GTK_BOX(main_vbox), info_frame, FALSE, TRUE, 0);
    gtk_widget_show(info_frame);

    info_table = gtk_table_new(ll, 2, FALSE);
    gtk_container_add (GTK_CONTAINER(info_frame), info_table);

    for (i = 0; i < ll; i++) {
      button = gtk_button_new_with_label (info_titles[i]);
      g_signal_connect (G_OBJECT (button), "clicked",
			G_CALLBACK(info_clicked), (gpointer)i);
      gtk_table_attach(GTK_TABLE (info_table), button, 0, 1, i, i+1,
		       GTK_SHRINK | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
      gtk_widget_show (button);

      info_labels[i] = gtk_label_new(files_in_use[i]);
      gtk_table_attach(GTK_TABLE (info_table), info_labels[i], 1, 2, i, i+1,
		       GTK_EXPAND |  GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
      gtk_misc_set_alignment(GTK_MISC(info_labels[i]), 0.0, 0.0);
      gtk_widget_show (info_labels[i]);
    }

    gtk_widget_show (info_table);
  }

  /* tool bar */

  {
    GdkVisual   * visual = gdk_visual_get_system();
    GdkColormap * colormap = gdk_colormap_new(visual, FALSE);
    GdkPixmap   * function_icon_pixmap  
	    = gdk_pixmap_colormap_create_from_xpm_d(
			    NULL,	/* window */
		   	    colormap,	/* colormap */
			    NULL,	/* mask */
			    NULL,	/* transparent color */
		 	   (gchar **)function_icon_h);
    GdkPixmap * file_icon_pixmap  
	    = gdk_pixmap_colormap_create_from_xpm_d(
			    NULL,	/* window */
		   	    colormap,	/* colormap */
			    NULL,	/* mask */
			    NULL,	/* transparent color */
		 	   (gchar **)file_icon_h);
    GdkPixmap * system_icon_pixmap  
	    = gdk_pixmap_colormap_create_from_xpm_d(
			    NULL,	/* window */
		   	    colormap,	/* colormap */
			    NULL,	/* mask */
			    NULL,	/* transparent color */
		 	   (gchar **)system_icon_h);
    GtkWidget * function_icon = gtk_image_new_from_pixmap(function_icon_pixmap, NULL);
    GtkWidget * system_icon = gtk_image_new_from_pixmap(system_icon_pixmap, NULL);
    GtkWidget * file_icon = gtk_image_new_from_pixmap(file_icon_pixmap, NULL);

    gtk_widget_show(function_icon);
    gtk_widget_show(system_icon);
    gtk_widget_show(file_icon);

    toolbar = gtk_toolbar_new();
    gtk_toolbar_set_orientation(GTK_TOOLBAR(toolbar), GTK_ORIENTATION_HORIZONTAL);
    gtk_toolbar_set_style(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_ICONS);
    gtk_widget_show(toolbar);

    gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),	/* toolbar 		*/
		            "Functions",		/* text			*/
			    "Show functions list.",	/* tooltip_text		*/
			    NULL,			/* tooltip_private_text	*/
			    function_icon,		/* icon			*/
			    G_CALLBACK(toolbar_callback),	/* callback	*/
			    (gpointer)TOOLBAR_FUNCTIONS);	/* user_data	*/

    gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),	/* toolbar 		*/
		            "File",			/* text			*/
			    "Show files list.",		/* tooltip_text		*/
			    NULL,			/* tooltip_private_text	*/
			    file_icon,		/* icon			*/
			    G_CALLBACK(toolbar_callback),	/* callback	*/
			    (gpointer)TOOLBAR_FILE);	/* user_data	*/

    gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),	/* toolbar 		*/
		            "System",			/* text			*/
			    "Show system hits.",		/* tooltip_text		*/
			    NULL,			/* tooltip_private_text	*/
			    system_icon,		/* icon			*/
			    G_CALLBACK(toolbar_callback),	/* callback	*/
			    (gpointer)TOOLBAR_SYSTEM);	/* user_data	*/

    gtk_box_pack_start(GTK_BOX(main_vbox), toolbar, FALSE, TRUE, 0);
  }


  /* status bar */

  {
    GtkWidget * status_frame;
    status_frame = gtk_frame_new(NULL);
    gtk_container_set_border_width(GTK_CONTAINER(status_frame), 5);
    gtk_box_pack_start(GTK_BOX(main_vbox), status_frame, FALSE, TRUE, 0);
    gtk_widget_show(status_frame);

    status_bar = gtk_label_new("");
    gtk_misc_set_alignment(GTK_MISC(status_bar), 0.0, 0.0);
    gtk_container_add(GTK_CONTAINER(status_frame), status_bar);
    gtk_widget_show(status_bar);
  }

  
  /* drawing area and rulers */
  
  {
    GtkWidget * table;
    GtkWidget * hrange;
    GtkWidget * left_frame;
    GtkWidget * right_frame;
    GtkWidget * scrollbar;

#define EVENT_METHOD(i, x) GTK_WIDGET_GET_CLASS(i)->x

    table = gtk_table_new( 3, 2, FALSE );
    gtk_box_pack_start(GTK_BOX(main_vbox), table, TRUE, TRUE, 0);

    drawing_area = gtk_drawing_area_new ();
    gtk_drawing_area_size (GTK_DRAWING_AREA (drawing_area),
			   drawing_area_width, drawing_area_height);
    gtk_table_attach( GTK_TABLE(table), drawing_area, 1, 2, 1, 2,
                      GTK_EXPAND|GTK_FILL, GTK_FILL, 0, 0 );
    gtk_widget_set_events(drawing_area, GDK_POINTER_MOTION_MASK |
			  GDK_POINTER_MOTION_HINT_MASK | GDK_BUTTON_PRESS_MASK);
    gtk_signal_connect (GTK_OBJECT (drawing_area), "expose_event",
			(GtkSignalFunc) expose_event, NULL);
    gtk_signal_connect (GTK_OBJECT(drawing_area),"configure_event",
			(GtkSignalFunc) configure_event, NULL);
    gtk_signal_connect (GTK_OBJECT (drawing_area), "button_press_event",
			(GtkSignalFunc)button_press_event, NULL);
    g_signal_connect(GTK_OBJECT(drawing_area), "motion_notify_event",
			       G_CALLBACK (motion_notify_event),
			       drawing_area);

    positions_history = malloc(POSITIONS_HISTORY_SLOTS  * sizeof(gdouble));
    bzero(positions_history, POSITIONS_HISTORY_SLOTS * sizeof(gdouble));



    
    hrange = gtk_hbox_new(FALSE, 1);
    gtk_table_attach( GTK_TABLE(table), hrange, 1, 2, 0, 1,
                      GTK_FILL, GTK_EXPAND|GTK_SHRINK|GTK_FILL, 0, 0 );
    gtk_widget_show(hrange);
    
    left_frame = gtk_frame_new(NULL);
    gtk_container_set_border_width(GTK_CONTAINER(left_frame), 5);
    left_label = gtk_label_new("|");
    gtk_widget_show(left_label);
    gtk_widget_show(left_frame);
    gtk_container_add(GTK_CONTAINER(left_frame), left_label);
    
    right_frame = gtk_frame_new(NULL);
    gtk_container_set_border_width(GTK_CONTAINER(right_frame), 5);
    right_label = gtk_label_new("|");
    gtk_widget_show(right_label);
    gtk_widget_show(right_frame);
    gtk_container_add(GTK_CONTAINER(right_frame), right_label);
    
    gtk_box_pack_start(GTK_BOX(hrange), left_frame,  FALSE, FALSE, 0);
    gtk_box_pack_end  (GTK_BOX(hrange), right_frame, FALSE, FALSE, 0);
    
    gtk_widget_show (hrange);



    
    vrule = gtk_vruler_new();
    gtk_ruler_set_metric( GTK_RULER(vrule), GTK_PIXELS );
    gtk_ruler_set_range( GTK_RULER(vrule), 0, 400, 10, 400 );
    g_signal_connect_swapped(GTK_OBJECT(drawing_area), "motion_notify_event",
                              G_CALLBACK (EVENT_METHOD (vrule, motion_notify_event)),
                              vrule);
    gtk_table_attach( GTK_TABLE(table), vrule, 0, 1, 1, 2,
                      GTK_FILL, GTK_EXPAND|GTK_SHRINK|GTK_FILL, 0, 0 );

    adjustment =  gtk_adjustment_new(0.25,	/* gdouble value		*/
				     0.0,	/* gdouble lower		*/
				     1.0,	/* gdouble upper		*/
				     0.01,	/* gdouble step_increment	*/
				     0.1,	/* gdouble page_increment	*/
				     1.0);	/* gdouble page_size		*/
    g_signal_connect(adjustment, "changed",
		     G_CALLBACK (adj_changed_event),
		     NULL);
    g_signal_connect(adjustment, "value-changed",
		     G_CALLBACK (adj_changed_event),
		     NULL);

    scrollbar = gtk_hscrollbar_new(GTK_ADJUSTMENT(adjustment));
    gtk_range_set_update_policy (GTK_RANGE (scrollbar),
                                 GTK_UPDATE_CONTINUOUS);
    gtk_table_attach( GTK_TABLE(table), scrollbar, 1, 2, 2, 3,
                      GTK_EXPAND|GTK_SHRINK|GTK_FILL, GTK_FILL, 0, 0 );

    gtk_widget_show (drawing_area);
    gtk_widget_show (vrule);
    gtk_widget_show (scrollbar);
    gtk_widget_show (table);
  }

  build_composite_tree();

  gtk_widget_show(main_window);

#ifdef DO_MEM_CHECK
  mtrace();
#endif
    
  gtk_main();

#ifdef DO_MEM_CHECK
  muntrace();
#endif
  
  if (positions_history) free(positions_history);
    
  return(0);
}

