/******************************************************************************
Copyright (C) 2002, 2003 Red Hat, Inc.

This file is part of op_visualise, a part of Oprofile.

This file may be redistributed and/or modifed under the terms of the GNU
General Public License as published by the Free Software Foundation; either
version 2, or (at your option) any later version.

op_visualise is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details. 

op_visualise was written, and is maintained, by Chris Moller,
Red Hat Inc, cmoller@redhat.com

This file provides functions that render oprofile data in various formats such
as Postscript and various image (e.g., GIF) and editable (e.g., FIG) formats.

******************************************************************************/

#ifdef ENABLE_PLOT
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <plot.h>
#define __USE_ISOC99 1
#include <math.h>

#include "op_visualise.h"
#include "print-modes.h"

#define ENABLE_ALABEL

#define G_MARGIN 0.1
#define NR_VALUE_TICS 10
/* NR_ADDR_TICS must be a multiple of 8 */
#define NR_ADDR_TICS 8

static gdouble addr_min;
static bfd_vma addr_min_l;
static bfd_vma addr_max_l;
static gdouble addr_span;
static gdouble value_max;

#define ADDR_XFORM(addr) (((double)addr - addr_min)/addr_span)
#define VALUE_XFORM(value) ((double)value/value_max)

extern GtkWidget * status_bar;
static GtkWidget * file_sel;

static print_format_e print_format = (print_format_e)0;
static print_format_e print_format_selected = (print_format_e)0;

static plPlotter *plotter;

static void
plot_callback(bfd_vma key,
	      unsigned int value)
{
  pl_fline_r(plotter,
	     ADDR_XFORM(key), 0.0,
	     ADDR_XFORM(key), VALUE_XFORM(value));
  pl_endpath_r(plotter);
}

static int
compute_majortic(double vm)
{
  int majortic;

  double ticfrac;
  double ticexp;
  double majorticmantissa;
  
  ticfrac = log10(vm/(double)NR_VALUE_TICS);
  ticfrac -= (ticexp = floor(ticfrac));
  majorticmantissa = ceil(exp(M_LN10 * ticfrac));
  if ((majorticmantissa > 2.0) && (majorticmantissa <= 5.0)) majorticmantissa = 5.0;
  else if (majorticmantissa > 5.0) majorticmantissa = 10.0;
  majortic = (int)lrint(majorticmantissa * exp(M_LN10 * ticexp));

  return majortic;
}

static void
draw_value_axis()
{
  pl_pencolorname_r (plotter, "blue");
  pl_fline_r(plotter, 0.09, 0.0, 0.09, 1.0);
  pl_endpath_r(plotter);

  {
    int majortic;
    int i,j,mt;
    char ttt[32];
    double vv;
    
    majortic = compute_majortic(value_max);
    for (mt = 0; mt < value_max; mt += majortic) {
      sprintf(ttt,"%d", mt);
      vv = VALUE_XFORM(mt);
      pl_fmove_r(plotter, 0.06, vv);
#ifdef ENABLE_ALABEL
      pl_alabel_r(plotter, 'r', 'c', ttt);
#endif
      pl_fline_r(plotter, 0.07, vv, 0.09, vv);
      pl_endpath_r(plotter);
      for (i = 0; i < 10; i++) {
	if ((j = mt + (i * majortic)/10) <= value_max) {
	  vv = VALUE_XFORM(j);
	  pl_fline_r(plotter, 0.08, vv, 0.09, vv);
	  pl_endpath_r(plotter);
	}
      }
    }
  }
}

static void
draw_addr_axis()
{
  unsigned int i,incr;
  double aa;
  char ttt[32];
#ifdef ENABLE_ALABEL
  char * sss;
  char * bbb;
#endif
  
  pl_pencolorname_r (plotter, "blue");
  pl_fline_r(plotter, 0.0, 0.1, 1.0, 0.1);
  pl_endpath_r(plotter);

#ifdef ENABLE_ALABEL
  asprintf(&sss,"Sample file: %s",files_in_use[SAMPLES_FILE]);
  asprintf(&bbb,"Binary file: %s",files_in_use[BINARY_FILE]);
  pl_fmove_r(plotter, 0.0, 0.04);
  pl_alabel_r(plotter, 'l', 't', sss);
  pl_endpath_r(plotter);
  pl_fmove_r(plotter, 0.0, 0.015);
  pl_alabel_r(plotter, 'l', 't', bbb);
  pl_endpath_r(plotter);
  free(sss);
  free(bbb);
#endif

  incr = (addr_max_l - addr_min_l)/NR_ADDR_TICS;
  for (i = addr_min_l; i <= addr_max_l; i+= incr) {
    aa = ADDR_XFORM((i /* - start_vma */));	/* fixme */
    if ((aa >= 0.0) && (aa <= 1.0)) {
      sprintf(ttt,"%08x",i);
      pl_fmove_r(plotter, aa, 0.07);
#ifdef ENABLE_ALABEL
      pl_alabel_r(plotter, 'c', 't', ttt);
#endif
      pl_fline_r(plotter, aa, 0.08, aa, 0.1);
      pl_endpath_r(plotter);
    }
  }
}

static int
export_graph(G_CONST_RETURN gchar * fn, gchar * type)
{
  plPlotterParams *plotter_params;
  FILE * pfile = NULL;
  int rc = 1;
  unsigned int min_l;
  unsigned int max_l;

  plotter_params = pl_newplparams ();
  pl_setplparam (plotter_params, "PAGESIZE", "letter");

  if (NULL == (pfile = fopen(fn, "w"))) {
    gtk_label_set(GTK_LABEL(status_bar), "Error opening plot file.");
    rc = 0;
  }
  
  if (rc && ((plotter = pl_newpl_r (type, NULL, pfile, stderr, plotter_params)) == NULL)) {
    gtk_label_set(GTK_LABEL(status_bar), "Couldn't create Plotter.");
    rc = 0;
  }

  if (rc && (pl_openpl_r (plotter) < 0)) {
    gtk_label_set(GTK_LABEL(status_bar), "Couldn't open Plotter.");
    rc = 0;
  }

  if (rc) {
    pl_fspace_r (plotter, 0.0, 0.0, 1.0, 1.0);
    pl_flinewidth_r (plotter, -1.0);
    pl_pencolorname_r (plotter, "black");
    pl_erase_r (plotter);

    pl_savestate_r(plotter);				/* x (addr) axis */
    pl_pencolorname_r (plotter, "green");
    pl_fsetmatrix_r(plotter,
		    0.9 - G_MARGIN,            0.0,
		    0.0,            0.9 - G_MARGIN,
		    G_MARGIN,                  0.0);
    pl_ffontsize_r(plotter, -1.0);
    draw_addr_axis();
    pl_endpath_r(plotter);
    pl_restorestate_r(plotter);

    pl_savestate_r(plotter);				/* y (value) axis */
    pl_fsetmatrix_r(plotter,
		    0.9 - G_MARGIN,            0.0,
		    0.0,            0.9 - G_MARGIN,
		    0.0,                  G_MARGIN);
    draw_value_axis();
    pl_restorestate_r(plotter);

    pl_savestate_r(plotter);				/* draw graph */
    pl_pencolorname_r (plotter, "black");
    pl_fsetmatrix_r(plotter,
		    0.9 - G_MARGIN,            0.0,
		    0.0,            0.9 - G_MARGIN,
		    G_MARGIN,             G_MARGIN);
    samples_travel(min_l = lrint(min_ruler_d) - 1,
		   max_l = lrint(max_ruler_d) + 1,
		   plot_callback, drawing_area->style->white_gc);
    
    if (symbol_list) {
      int i;
      unsigned long vv;
      double dvv;
      
      pl_pencolorname_r (plotter, "red");
      for ( i = 0; i < sym_count; i++) {
	vv = symbol_list[i]->value /* - start_vma */;  /* fixme */
	dvv = ADDR_XFORM(vv);
	if ((min_l <= vv) && (vv <= max_l)) {
	  pl_fmove_r(plotter, dvv, (i & 1) ? 1.075 : 1.05);
#ifdef ENABLE_ALABEL
	  pl_alabel_r(plotter, 'c', 'x', symbol_list[i]->name);
#endif
	  pl_fline_r(plotter, dvv, 1.0, dvv, 1.045);
	  pl_endpath_r(plotter);
	}
      }
    }
    pl_restorestate_r(plotter);
    
    if (pl_closepl_r (plotter) < 0) {
      gtk_label_set(GTK_LABEL(status_bar), "Couldn't close Plotter.");
      rc = 0;
    }
     
    if (rc && (pl_deletepl_r (plotter) < 0)) {
      gtk_label_set(GTK_LABEL(status_bar), "Couldn't delete Plotter.");
      rc = 0;
    }
  }
  if (pfile) fclose(pfile);
  return rc;
}

static void
file_select_ok (
		GtkObject * object __attribute__ ((unused)),
		gint    data)
{
  gdouble addr_max;
  
  G_CONST_RETURN gchar * fn = gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_sel));

  if ('*' != *fn) {
    addr_min = min_ruler_d;
    addr_max = max_ruler_d;
    addr_span = addr_max - addr_min;
    addr_min_l = (bfd_vma)(~15 & lrint(addr_min));
    addr_max_l = (bfd_vma)(~15 & (15 + lrint(addr_max)));
    gtk_ruler_get_range(GTK_RULER(vrule), &value_max, NULL, NULL, NULL);

    print_format = print_format_selected;
    if (export_graph(fn, print_formats[print_format].type)) {
      gtk_label_set(GTK_LABEL(status_bar), "");
      gtk_widget_destroy(file_sel);
      window_open &= ~(1 << data);
    }
  }
  else gtk_label_set(GTK_LABEL(status_bar), "Invalid plt file name.");
}

static void
file_select_cancel (
		    GtkObject * object __attribute__ ((unused)),
		    gint        data)
{
  gtk_widget_destroy(file_sel);
  window_open &= ~(1 << data);
}

static void
print_option_callback( GtkWidget *item __attribute__ ((unused)),
		       print_format_e pf)
{
  print_format_selected = pf;
  gtk_file_selection_complete(GTK_FILE_SELECTION(file_sel),
			      print_formats[pf].extension);
}


static GtkWidget *
make_menu_item (gchar     *name,
		GCallback  callback,
		gpointer   data)
{
    GtkWidget *item;

    item = gtk_menu_item_new_with_label (name);
    g_signal_connect (G_OBJECT (item), "activate",
                      callback, data);
    gtk_widget_show (item);

    return item;
}

void
handle_print_file(GtkWidget * w __attribute__ ((unused)),
		  gint data,
		  void * callback_action __attribute__ ((unused)))
{
  char * wname;
  GtkWidget * save_options;

  if (!(window_open & (1 << data))) {
    window_open |= (1 << data);
    
    wname = "Print File";

    file_sel = gtk_file_selection_new(wname);
    gtk_signal_connect(GTK_OBJECT(file_sel),
		     "destroy",
		     GTK_SIGNAL_FUNC(file_select_cancel),
		     (gpointer)data);
    gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(file_sel));
    gtk_signal_connect (GTK_OBJECT(GTK_FILE_SELECTION(file_sel)->ok_button), "clicked",
			GTK_SIGNAL_FUNC (file_select_ok),
			(gpointer)data);
    gtk_signal_connect (GTK_OBJECT(GTK_FILE_SELECTION(file_sel)->cancel_button), "clicked",
			GTK_SIGNAL_FUNC (file_select_cancel),
			(gpointer)data);

    {		/* from gimp app/fileops.c:561	*/
      GtkWidget *frame;
      GtkWidget *hbox;
      GtkWidget *label;
      GtkWidget *option_menu;

      save_options = gtk_hbox_new (TRUE, 1);

      frame = gtk_frame_new ("Save Options");
      gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
      gtk_box_pack_start (GTK_BOX (save_options), frame, TRUE, TRUE, 4);

      hbox = gtk_hbox_new (FALSE, 4);
      gtk_container_set_border_width (GTK_CONTAINER (hbox), 4);
      gtk_container_add (GTK_CONTAINER (frame), hbox);
      gtk_widget_show (hbox);

      label = gtk_label_new ("Format:");
      gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
      gtk_widget_show (label);

      option_menu = gtk_option_menu_new ();
      gtk_box_pack_start (GTK_BOX (hbox), option_menu, TRUE, TRUE, 0);
      gtk_widget_show (option_menu);


      {
	GtkWidget * item;
	GtkWidget * top_item = NULL;
	GtkWidget * menu;
	unsigned int i;

	menu = gtk_menu_new();

	for (i = 0; i < sizeof(print_formats)/sizeof(print_format_s); i++) {
	  item = make_menu_item (print_formats[i].name, G_CALLBACK (print_option_callback),
				 GINT_TO_POINTER(i));
	  gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
	  if (i == print_format) top_item = item;
	}

	if (top_item) gtk_menu_reorder_child(GTK_MENU(menu), top_item, 0);

	gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), menu);
      }
      
      gtk_widget_show (frame);
      gtk_widget_show (save_options);

      /* pack the containing save_options hbox into the save-dialog */
      gtk_box_pack_end (GTK_BOX (GTK_FILE_SELECTION (file_sel)->main_vbox),
                        save_options, FALSE, FALSE, 0);
    }
    
    print_format_selected = print_format;
    
    gtk_widget_show(file_sel);
    gtk_file_selection_complete(GTK_FILE_SELECTION(file_sel),
				print_formats[print_format].extension);
  }
}

#endif





