/*****************************************************************************
Copyright (C) 2002, 2003 Red Hat, Inc.

This file is part of op_visualise, a part of Oprofile.

This file may be redistributed and/or modifed under the terms of the GNU
General Public License as published by the Free Software Foundation; either
version 2, or (at your option) any later version.

op_visualise is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

op_visualise was written, and is maintained, by Chris Moller,
Red Hat Inc, cmoller@redhat.com

This file provides both standard and custom sample file selection functions

******************************************************************************/

  /* #define SHOW_SAMPLES_TREE */

#define __USE_GNU	/* enables GNU C extensions */
#define _GNU_SOURCE	/* enables GNU C extensions */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <regex.h>

#include "op_visualise.h"

extern GtkWidget * status_bar;

static GtkWidget * sample_list_window = 0;	/* sample file main window		*/
static GtkWidget * files_tree;			/* samples dir file names tree view	*/
static GtkWidget * dir_tree;			/* samples dir subdir tree view		*/
static GtkWidget * history_options_menu;

/* samples node flags */

#define SAMPLES_NODE_ROOT	0		/* root					  */
#define SAMPLES_NODE_INTERNAL	1		/* internal (non-leaf) node (a directory) */
#define SAMPLES_NODE_LEAF	2		/* leaf node (regular file)		*/
#define SAMPLES_NODE_HIDDEN	4		/* "redundant" leaf node		*/

#define IS_SAMPLES_LEAF(v)	((SAMPLES_NODE_LEAF & (v)))
#define IS_SAMPLES_HIDDEN(v)	((SAMPLES_NODE_HIDDEN & (v)))

typedef struct _samples_tree_s {
  char * aggregate_b;				/* aggregate binary file name path	*/
  char * aggregate;				/* aggregate sample file name path	*/
  char * element;				/* file name element			*/
  struct _samples_tree_s ** children;		/* children table root			*/
  int children_next;				/* nr etys in children table		*/
  int children_max;				/* nr etys alloced in children table	*/
  struct _samples_tree_s * parent;		/* parent				*/
  int    flags;					/* node flags				*/
} samples_tree_s;				/* samples tree node structure		*/
#define SAMPLES_TREE_CHILD_INCR 8		/* children table realloc increment	*/

#ifndef DONT_USE_OBSTACKS
static struct obstack samples_tree_obstack;		/* sample tree nodes		     */
static struct obstack samples_tree_element_obstack;	/* samples tree name element strings */
#endif
samples_tree_s * samples_tree_root;		/* root of sample tree nodes		*/

#ifndef DONT_USE_OBSTACKS
static struct obstack history_obstack;		/* history strings			*/
#endif
static char ** history_ptrs = 0;		/* history strings ptr table root	*/
static int history_ptrs_next = 0;		/* nr etys in history strings table	*/
static int history_ptrs_max = 0;		/* nr etys alloced in history table	*/
#define HISTORY_PTRS_INCR 8			/* realloc increment for history table	*/
static int current_history_idx;			/* idx of current history selection	*/
static int * history_idxs_sorted = 0;
static int history_idxs_sorted_valid = 0;

#ifndef DONT_USE_OBSTACKS
static struct obstack dir_obstack;		/* samples dir subdir name strings	*/
static struct obstack dir_ptr_obstack;		/* samples dir subdir name ptr table	*/
#endif
static int dir_count;				/* nr subdirs in samples dir		*/
static char ** dir_list_ptrs;			/* samples dir subdir name ptr table root */

static int file_to_replace;

static GtkWidget * progress_bar_dialog;
static GtkWidget * progress_bar;
static double progress_bar_max;

static gboolean progress_rc;

static int which_dir;
static long int global_which;

static GtkWidget * hit_list_window;

#ifdef USE_OLD_FASHIONED_SELECT
static char * selected_fn = 0;
#else
static samples_tree_s * selected_samples_tree = 0;
#endif

static char * dialog_fn = 0;

static void update_samples_file(long int which, samples_tree_s * st);

/*									*/
/*									*/
/*	progress bar stuff   						*/
/*									*/
/*									*/

/*									*/
/*	update the progress bar and wait until the event queue is empty	*/
/*									*/

gboolean
update_progress_bar(double tp)
{
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress_bar), tp/progress_bar_max);
  while (gtk_events_pending()) gtk_main_iteration();
  return progress_rc;
}


/*									*/
/*	react to the progress bar cancel button				*/
/*									*/

static void
progress_bar_response (GtkDialog *dialog __attribute__ ((unused)),
                       gint arg1,
                       gpointer user_data __attribute__ ((unused)))
{
  if (GTK_RESPONSE_CANCEL == arg1) {
    progress_rc = TRUE;
  }
}


/*									*/
/*	create the progress bar						*/
/*									*/

void
initialise_progress_bar(double pbm)
{
  progress_bar_dialog = gtk_dialog_new_with_buttons("Reading symbols...",
						    GTK_WINDOW(main_window),
						    GTK_DIALOG_DESTROY_WITH_PARENT,
						    GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
						    NULL);

  g_signal_connect (GTK_OBJECT (progress_bar_dialog), 
		    "response", 
		    G_CALLBACK (progress_bar_response),
		    GTK_OBJECT (progress_bar_dialog));


  progress_bar = gtk_progress_bar_new();

  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(progress_bar_dialog)->vbox),
		    progress_bar);
  progress_bar_max = pbm;
  progress_rc = FALSE;
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress_bar), 0.0);
  gtk_widget_show(progress_bar);
  gtk_widget_show(progress_bar_dialog);
  while (gtk_events_pending()) gtk_main_iteration();
}


/*									*/
/*	destroy the progress bar					*/
/*									*/

void
remove_progress_bar(void)
{
  gtk_widget_destroy(progress_bar_dialog);
}


/*								*/
/*								*/
/*								*/
/*	SAMPLES TREE ROUTINES					*/
/*								*/
/*								*/
/*								*/

#ifdef SHOW_SAMPLES_TREE
/*						*/
/* traverse and display the samples tree	*/
/*						*/

static void
show_samples_tree(int indent, samples_tree_s * samples_tree_current)
{
  int i;

  if (samples_tree_current) {
    fprintf(stderr,"  aggregate = %*s%s\n", indent, " ", samples_tree_current->aggregate);
    fprintf(stderr,"aggregate_b = %*s%s\n", indent, " ", samples_tree_current->aggregate_b);
    fprintf(stderr,"  element   = %*s%s\n", indent, " ", samples_tree_current->element);
    for (i = 0; i < samples_tree_current->children_next; i++)
      show_samples_tree(indent+2, samples_tree_current->children[i]);
  }
}
#endif


/*								*/
/* traverse and free malloced portions of the samples tree	*/
/*								*/

static void
free_samples_tree(samples_tree_s * samples_tree_current)
{
  int i;
  
  if (samples_tree_current) {
    for (i = 0; i < samples_tree_current->children_next; i++)
      free_samples_tree(samples_tree_current->children[i]);
    if (samples_tree_current->children) free(samples_tree_current->children);
    if (samples_tree_current->aggregate)   free(samples_tree_current->aggregate);
    if (samples_tree_current->aggregate_b) free(samples_tree_current->aggregate_b);
  }
}
     

/*								*/
/* append an entry to the current samples tree node child table	*/
/*								*/

static samples_tree_s *
append_child(samples_tree_s * samples_tree_current,
	     char * ptr,
	     int snt)
{
  samples_tree_s * samples_tree_child;
  char * ep;

#ifndef DONT_USE_OBSTACKS
  samples_tree_child = obstack_alloc(&samples_tree_obstack, sizeof(samples_tree_s));
#endif
  
  asprintf(&samples_tree_child->aggregate, "%s}%s",
	   strcmp(samples_tree_current->aggregate, "}") ? samples_tree_current->aggregate : "",
	   ptr);
  
  if (BINARIES_DIR == which_dir)
    if (IS_SAMPLES_LEAF(snt) && (ep = strchr(ptr, '#'))) *ep = '\0';
  
#ifndef DONT_USE_OBSTACKS
  samples_tree_child->element   = obstack_copy0(&samples_tree_element_obstack, ptr, strlen(ptr));
#endif
  
  if (BINARIES_DIR != which_dir)
    if (IS_SAMPLES_LEAF(snt) && (ep = strchr(ptr, '#'))) *ep = '\0';
  
  asprintf(&samples_tree_child->aggregate_b, "%s/%s",
	   strcmp(samples_tree_current->aggregate_b, "/") ? samples_tree_current->aggregate_b : "",
	   ptr);
  
  samples_tree_child->children      = 0;
  samples_tree_child->children_next = 0;
  samples_tree_child->children_max  = 0;
  samples_tree_child->parent  	    = samples_tree_current;
  samples_tree_child->flags  	    = snt;

  if (samples_tree_current->children_next >= samples_tree_current->children_max) {
    samples_tree_current->children_max += SAMPLES_TREE_CHILD_INCR;
    samples_tree_current->children
      = realloc(samples_tree_current->children,
		samples_tree_current->children_max * sizeof(samples_tree_s *));
  }	
  samples_tree_current->children[samples_tree_current->children_next++]
    = samples_tree_child;
  return samples_tree_child;
}

/*								*/
/* build the samples tree					*/
/*								*/
/*	this fcn scans dir containing the sample files and:	*/
/*								*/
/*	parses the '}' delimited sample file names to create	*/
/*	an N-ary tree of the file name elements that maps to	*/
/*	the executable files that correspond to those samples	*/
/*								*/
/*	creates a list of the subdirs under the samples dir	*/
/*								*/
/*	results:								*/
/*		samples_tree_s * samples_tree_root 	-- the samples tree		*/
/*		int dir_count				-- the number directory entries	*/
/*		char ** dir_list_ptrs			-- a list of ptr to the dir names	*/

#define SAMPLE_TREE_TOP_STRING "<top>"

static void
build_samples_tree (char * samp_dir, char * bin_dir)
{
  struct dirent * dir;
  DIR * samples_dir;
  DIR * binaries_dir;
  char * cpy;
  char * ptr;
  char * ptr_e;
  int    cnt;
  samples_tree_s * samples_tree_current;
  samples_tree_s * samples_tree_child;
  int i;
  
  if (samp_dir) {
    if (!(samples_dir = opendir(samp_dir))) 
      gtk_label_set(GTK_LABEL(status_bar), "Missing samples directory.");
    else {
      samples_tree_root = NULL;

      while ((dir = readdir(samples_dir))) {
	cpy = strdupa(dir->d_name);

	switch (dir->d_type) {
	case DT_REG:
	  if (*cpy == '}') {
	    int snt;
	    if (!samples_tree_root) {
#ifndef DONT_USE_OBSTACKS
	      samples_tree_root = obstack_alloc(&samples_tree_obstack, sizeof(samples_tree_s));
	      samples_tree_root->element = obstack_copy0(&samples_tree_element_obstack,
							 SAMPLE_TREE_TOP_STRING,
							 strlen(SAMPLE_TREE_TOP_STRING));
#endif
	      samples_tree_root->aggregate     = strdup("}");
	      samples_tree_root->aggregate_b   = strdup("/");
	      samples_tree_root->children      = 0;
	      samples_tree_root->children_next = 0;
	      samples_tree_root->children_max  = 0;
	      samples_tree_root->parent  	     = NULL;
	      samples_tree_root->flags  	     = SAMPLES_NODE_ROOT;
	    }
	    cnt = -1;
	    if ((ptr = strchr(cpy, '#'))) cnt = atoi(ptr);
	
	    samples_tree_current = samples_tree_root;
	
	    /* scan the file name looking for '}' separators.  append name elements to the tree */
	
	    for (ptr_e = ptr = cpy + 1; ptr_e; ) {
	      if ((ptr_e = strchr(ptr, '}'))) {
		*ptr_e++ = '\0';

	      /* check if the element already exists.  if so, follow that branch */
		samples_tree_child = 0;
		for (i = 0; i < samples_tree_current->children_next; i++) {
		  if (!strcmp(ptr, samples_tree_current->children[i]->element)) {
		    samples_tree_child = samples_tree_current->children[i];
		    break;
		  }
		}

	    /* if not already an element, append it. */
		if (!samples_tree_child)
		  samples_tree_child = append_child(samples_tree_current, ptr, SAMPLES_NODE_INTERNAL);
		samples_tree_current = samples_tree_child;
	    
		ptr = ptr_e;
	      }
	    }
	    /* final element is a leaf */
	    snt = SAMPLES_NODE_LEAF;
	    if (BINARIES_DIR == which_dir) {
	      for (i = 0; i < samples_tree_current->children_next; i++) {
		if (!strncmp(ptr, samples_tree_current->children[i]->element,
			     strlen(samples_tree_current->children[i]->element))) {
		  snt |= SAMPLES_NODE_HIDDEN;
		  break;
		}
	      }
	    }
	    append_child(samples_tree_current, ptr, snt);
	  }
	  break;
	case DT_DIR:
	  if (SAMPLES_DIR == which_dir) {
	    if (strcmp(cpy,"."))
#ifndef DONT_USE_OBSTACKS
	      obstack_ptr_grow(&dir_ptr_obstack, obstack_copy0 (&dir_obstack, cpy, strlen(cpy)));
#endif			
	  }
	  break;
	}
      }
      closedir(samples_dir);
    }
  }

  if (bin_dir) {
    if (!(binaries_dir = opendir(bin_dir))) 
      gtk_label_set(GTK_LABEL(status_bar), "Missing binaries directory.");
    else {
      while ((dir = readdir(binaries_dir))) {
	cpy = strdupa(dir->d_name);
	switch (dir->d_type) {
	case DT_REG:
	  break;
	case DT_DIR:
	  if (strcmp(cpy,"."))
#ifndef DONT_USE_OBSTACKS
	    obstack_ptr_grow(&dir_ptr_obstack, obstack_copy0 (&dir_obstack, cpy, strlen(cpy)));
#endif			
	  break;
	}
      }
      closedir(binaries_dir);
    }
  }
  
#ifndef DONT_USE_OBSTACKS
  dir_count = obstack_object_size (&dir_ptr_obstack)/sizeof(char *);
  dir_list_ptrs = obstack_finish(&dir_ptr_obstack);
#endif
  
#ifdef SHOW_SAMPLES_TREE
  show_samples_tree(0, samples_tree_root);
#endif
}

/*									*/
/* recursively store the samples tree elements in the tree view model	*/
/*									*/

#define VALID_ITER(iter, list_store) (iter!= NULL && iter->user_data != NULL && list_store->stamp == iter->stamp)

static void
populate_tree_model (GtkTreeStore *store, GtkTreeIter * iter1, samples_tree_s * bt)
{
  GtkTreeIter iter2;
  int i;

  if ((iter1 != NULL) && !VALID_ITER(iter1, store)) {
    fprintf(stderr, "Hey!, stop!\n");
  }
  if (!IS_SAMPLES_HIDDEN(bt->flags)) {
    gtk_tree_store_append(store, &iter2, iter1);
    gtk_tree_store_set(store, &iter2,
		       0, bt->element,
		       -1);
    for (i = 0; i < bt->children_next; i++)
      populate_tree_model(store, &iter2, bt->children[i]);		     
  }
}






/*								*/
/*								*/
/*								*/
/*	HISTORY ROUTINES					*/
/*								*/
/*								*/
/*								*/

static void
set_subdirs_list(int dlc, char ** dlp)
{
  GtkListStore *model;
  GtkTreeIter iter;
  int i;
  
  if (dlc > 0) {
    model = gtk_list_store_new (1, G_TYPE_STRING);
    gtk_tree_view_set_model (GTK_TREE_VIEW (dir_tree), GTK_TREE_MODEL (model));
    for (i = 0; i < dlc; i++) {
      gtk_list_store_append (GTK_LIST_STORE (model), &iter);
      gtk_list_store_set (GTK_LIST_STORE (model),
			  &iter,
			  0, dlp[i],
			  -1);
    }
    g_object_unref (G_OBJECT (model));
  }
  else
    gtk_tree_view_set_model (GTK_TREE_VIEW (dir_tree), NULL);
}

static void
set_files_tree_list()
{
  if (samples_tree_root) {
    GtkTreeStore * store = gtk_tree_store_new (1, G_TYPE_STRING);
    populate_tree_model(store, NULL, samples_tree_root);

    gtk_tree_view_set_model(GTK_TREE_VIEW(files_tree), GTK_TREE_MODEL (store));

    g_object_unref (G_OBJECT (store));
  }
  else 
    gtk_tree_view_set_model(GTK_TREE_VIEW(files_tree), NULL);
}

static void
rebuild_dir_info(char * path)
{
  if (SAMPLES_DIR == which_dir) {
    free_samples_tree(samples_tree_root);
#ifndef DONT_USE_OBSTACKS
    obstack_free(&samples_tree_obstack, NULL);
    obstack_free(&samples_tree_element_obstack, NULL);
    obstack_init(&samples_tree_obstack);
    obstack_init(&samples_tree_element_obstack);
#endif
  }
#ifndef DONT_USE_OBSTACKS
  obstack_free(&dir_obstack, NULL);
  obstack_free(&dir_ptr_obstack, NULL);
  obstack_init(&dir_obstack);
  obstack_init(&dir_ptr_obstack);
#endif

  switch (which_dir) {
  case SAMPLES_DIR:
    build_samples_tree(path, NULL);
    set_files_tree_list();
    break;
  case BINARIES_DIR:
    build_samples_tree(NULL, path);
    break;
  }
  
  set_subdirs_list(dir_count, dir_list_ptrs);
}

/*									*/
/* callback for history selection					*/
/*									*/

static void
history_selection_callback(GtkWidget *item __attribute__ ((unused)), 
			   gpointer ppath_idx)
{
  current_history_idx = *((int *)(ppath_idx));
  rebuild_dir_info(history_ptrs[current_history_idx]);
}


/*									*/
/* create an item in the history list					*/
/*									*/

static GtkWidget *
make_history_menu_item (gchar     *name,
			GCallback  callback,
			gpointer   data)
{
    GtkWidget *item;

    item = gtk_menu_item_new_with_label (name);
    g_signal_connect (G_OBJECT (item), "activate",
                      callback, data);
    gtk_widget_show (item);

    return item;
}

/*									*/
/* catenate an entry to the history list				*/
/*									*/

static void
cat_path(char * path)
{
  if (history_ptrs_next >= history_ptrs_max) {
    history_ptrs_max += HISTORY_PTRS_INCR;
    history_ptrs = realloc(history_ptrs, history_ptrs_max * sizeof(char *));
  }
  history_ptrs[history_ptrs_next++] = path;
  history_idxs_sorted_valid = 0;
}


/*									*/
/* create the history table						*/
/*									*/

static void
build_history_table()
{
  char * path;
  unsigned int i;
  char * sdir = files_in_use[which_dir];

#ifndef DONT_USE_OBSTACKS
  obstack_init(&history_obstack);
  obstack_init(&dir_obstack);
  obstack_init(&dir_ptr_obstack);
  obstack_init(&samples_tree_obstack);
  obstack_init(&samples_tree_element_obstack);
#endif

  if (sdir) {
    char * cp;
    char * wfn = strdupa(sdir);
    current_history_idx = 0;

    for (i = 0; ; i++) {
#ifndef DONT_USE_OBSTACKS
      path = obstack_copy0 (&history_obstack, wfn, strlen(wfn));
#endif
      cat_path(path);
      
      if ('\0' == *(wfn+1)) break;
      if ((cp = strrchr(wfn, '/'))) *(cp + ((cp == wfn) ? 1 : 0)) = '\0';
    }
  }
  else {
#ifndef DONT_USE_OBSTACKS
    path = obstack_copy0 (&history_obstack, "/", 1);
#endif
    cat_path(path);
  }
}


/*									*/
/* qsort/bsearch compare functions					*/
/*									*/

static int
history_idxs_sort_fcn(const void * a, const void * b)
{
  return (strcmp(history_ptrs[*(int *)a], history_ptrs[*(int *)b]));
}

static int
history_idxs_search_fcn(const void * a, const void * b)
{
  return (strcmp((char *)a, history_ptrs[*(int *)b]));
}

static int *
find_history(char * new_dir)
{
  int i;
  
  if (!history_idxs_sorted_valid) {
    history_idxs_sorted = realloc(history_idxs_sorted, history_ptrs_next * sizeof(int));
    for (i = 0; i < history_ptrs_next; i++) history_idxs_sorted[i] = i;
    qsort(history_idxs_sorted, history_ptrs_next, sizeof(int), history_idxs_sort_fcn);
    history_idxs_sorted_valid = 1;
  }
  return bsearch(new_dir, history_idxs_sorted, history_ptrs_next, sizeof(int),
		 history_idxs_search_fcn);
}


/*									*/
/* store ety in the history table					*/
/*									*/

static void
populate_history_menu()
{
  GtkWidget * item;
  GtkWidget * menu = gtk_menu_new ();
  long int i;
	  
  for (i = 0; i < (long int)history_ptrs_next; i++) {
    item = make_history_menu_item (history_ptrs[i],
				   G_CALLBACK (history_selection_callback), 
				   (gpointer)i);
    gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
  }

  gtk_option_menu_set_menu (GTK_OPTION_MENU (history_options_menu), menu);
}




/*								*/
/*								*/
/*								*/
/*	SUBDIRS ROUTINES					*/
/*								*/
/*								*/
/*								*/


/*									*/
/* subdir selection callback						*/
/*									*/

static void
subdir_activated(GtkTreeView * treeview __attribute__ ((unused)),
		 GtkTreePath * arg1,
		 GtkTreeViewColumn * arg2 __attribute__ ((unused)),
		 gpointer data __attribute__ ((unused)))
{
  char * new_dir;
  int * hit;
  int * path = gtk_tree_path_get_indices(arg1);
  
  if (path[0] == 0) {		/* .. (parent) */
    char * pt_ptr;
    new_dir = strdupa(history_ptrs[current_history_idx]);
    if ((pt_ptr = strrchr(new_dir+1, '/'))) *pt_ptr = '\0';
  }
  else asprintf(&new_dir,"%s/%s",
		strcmp("/", history_ptrs[current_history_idx]) ? history_ptrs[current_history_idx] : "",
		dir_list_ptrs[path[0]]);

  if ((hit = find_history(new_dir))) {
    if (current_history_idx != *hit) {
      current_history_idx = *hit;
      gtk_option_menu_set_history(GTK_OPTION_MENU(history_options_menu), current_history_idx);
      rebuild_dir_info(history_ptrs[current_history_idx]);
    }
  }
  else {
    cat_path(new_dir);
    gtk_option_menu_remove_menu(GTK_OPTION_MENU(history_options_menu));
    populate_history_menu();
    current_history_idx = history_ptrs_next - 1;
    gtk_option_menu_set_history(GTK_OPTION_MENU(history_options_menu), current_history_idx);
    rebuild_dir_info(history_ptrs[current_history_idx]);
  }
  gtk_tree_view_expand_all(GTK_TREE_VIEW(files_tree));
}


/*									*/
/* create the samples dir subdirs gtk list				*/
/*									*/

static GtkWidget *
create_subdirs_list(int dlc, char ** dlp)
{

  GtkWidget *scrolled_window;
  GtkCellRenderer *cell;
  GtkTreeViewColumn *column;


  scrolled_window = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
				  GTK_POLICY_AUTOMATIC,
				  GTK_POLICY_AUTOMATIC);

  dir_tree = gtk_tree_view_new ();
  gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrolled_window),
					 dir_tree);
  gtk_widget_show (dir_tree);
  gtk_signal_connect (GTK_OBJECT(dir_tree), "row-activated",
		      GTK_SIGNAL_FUNC (subdir_activated),
		      NULL);

  set_subdirs_list(dlc, dlp);
  
  cell = gtk_cell_renderer_text_new ();

  column = gtk_tree_view_column_new_with_attributes ("Directories",
						     cell,
						     "text", 0,
						     NULL);

  gtk_tree_view_append_column (GTK_TREE_VIEW (dir_tree),
			       GTK_TREE_VIEW_COLUMN (column));

  return scrolled_window;
}

static void
build_button_box(GtkWidget * vbox,
		 long int which,
		 void (*accept_clicked)(),
		 void (*cancel_clicked)())
{
  GtkWidget * frame  = gtk_frame_new(NULL);
  GtkWidget * shbox  = gtk_hbox_new(TRUE, 2);
  GtkWidget * accept = gtk_button_new_from_stock(GTK_STOCK_OK);
  GtkWidget * cancel = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
      
  gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 0);
  gtk_widget_show(shbox);

  gtk_signal_connect (GTK_OBJECT(accept), "clicked",
		      GTK_SIGNAL_FUNC (accept_clicked),
		      (gpointer)which);

  gtk_signal_connect (GTK_OBJECT(cancel), "clicked",
		      GTK_SIGNAL_FUNC (cancel_clicked),
		      (gpointer)which);
      
      
  gtk_box_pack_end (GTK_BOX (shbox), cancel, FALSE, FALSE, 0);
  gtk_widget_show (cancel);
      
  gtk_box_pack_end (GTK_BOX (shbox), accept, TRUE, FALSE, 0);
  gtk_widget_show (accept);
      
  gtk_container_add (GTK_CONTAINER (frame), shbox);
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);
 
  gtk_widget_show(frame);
}


/*									*/
/* destroy the top-level samples selection menu				*/
/*									*/

static void
sample_list_cancel(GtkObject * object __attribute__ ((unused)),
		   gpointer    pwhich)
{
  free_samples_tree(samples_tree_root);
  samples_tree_root = NULL;
  if (history_ptrs) free(history_ptrs);
  history_ptrs = NULL;
  history_ptrs_next = history_ptrs_max = 0;
  if (history_idxs_sorted) free(history_idxs_sorted);
  history_idxs_sorted = NULL;
#ifndef DONT_USE_OBSTACKS
  obstack_free(&history_obstack, NULL);
  obstack_free(&samples_tree_obstack, NULL);
  obstack_free(&samples_tree_element_obstack, NULL);
  obstack_free(&dir_obstack, NULL);
  obstack_free(&dir_ptr_obstack, NULL);
#endif
  gtk_label_set(GTK_LABEL(status_bar), "");
  window_open &= ~(1 << (long int)pwhich);
}


/*									*/
/* accept/open the suggested binary					*/
/*									*/

static void
dialog_callback(GtkDialog * dialog __attribute__ ((unused)),
		gint arg1,
		gchar * fn __attribute__ ((unused)))
{
  char * kk;
  
  if (GTK_RESPONSE_YES == arg1) {
    if (!files_in_use[file_to_replace] || strcmp(dialog_fn, files_in_use[file_to_replace])) {
      if (files_in_use[file_to_replace]) free(files_in_use[file_to_replace]);
      files_in_use[file_to_replace] = strdup(dialog_fn);
      kk = files_in_use[file_to_replace] + (strlen(files_in_use[file_to_replace]) - 1);
      if (('/' == *kk) && ('\0' != *(kk+1))) *kk = '\0';
      file_changed |= (1 << file_to_replace);
      gtk_label_set(GTK_LABEL(info_labels[file_to_replace]), files_in_use[file_to_replace]);
    }
  }
}


/*									*/
/* pop up dialog to suggest new binary					*/
/*									*/

static void
open_dialog(char * mstring)
{
  GtkWidget * dialog;
  
  dialog = gtk_message_dialog_new(GTK_WINDOW(main_window),
				  GTK_DIALOG_DESTROY_WITH_PARENT,
				  GTK_MESSAGE_QUESTION,
				  GTK_BUTTONS_YES_NO,
				  mstring, dialog_fn);
  g_signal_connect_swapped (GTK_OBJECT (dialog), 
			    "response", 
			    G_CALLBACK (dialog_callback),
			    GTK_OBJECT (dialog));
  gtk_dialog_run (GTK_DIALOG (dialog));
  gtk_widget_destroy (dialog);
}


/*									*/
/* file selection callback						*/
/*									*/

static void
suggestion_activated (GtkTreeView * treeview __attribute__ ((unused)),
		      GtkTreePath * arg1 __attribute__ ((unused)),
		      GtkTreeViewColumn * arg2 __attribute__ ((unused)),
		      gpointer which __attribute__ ((unused)))
{
  if (dialog_fn) {
    file_to_replace = SAMPLES_FILE;
    dialog_callback(NULL, GTK_RESPONSE_YES, NULL);
    gtk_widget_destroy(hit_list_window);
    free(dialog_fn);
    dialog_fn = 0;;
    maybe_visualise_data();
  }
}

static void
suggestion_sel_changed(GtkTreeSelection * selection,
		       gpointer data __attribute__ ((unused)))
{
  GtkTreeIter iter;
  GtkTreeModel * model;

  if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
    gtk_tree_model_get(model, &iter, 0, &dialog_fn, -1);
  }
}

static void
suggestion_cancel_clicked (GtkObject * object __attribute__ ((unused)),
			   gint       data __attribute__ ((unused)))
{
  gtk_widget_destroy(hit_list_window);
  if (dialog_fn) free(dialog_fn);
  dialog_fn = 0;
}

static void
suggestion_accept_clicked (GtkButton * button __attribute__ ((unused)),
			   gpointer data __attribute__ ((unused)))
{
  suggestion_activated(NULL, NULL, NULL, (gpointer)global_which);
}

static void
update_samples_file(long int lwhich, samples_tree_s * st)
{
  int which = (int)lwhich;
  char * pn;
  char * qn;
  char * dn;

  if (SAMPLES_FILE == which) 
    asprintf(&pn, "%s/%s",
	     strcmp("/", files_in_use[SAMPLES_DIR]) ? files_in_use[SAMPLES_DIR] : "",
	     st->aggregate);
  else
    asprintf(&pn, "%s%s",
	     strcmp("/", files_in_use[BINARIES_DIR]) ? files_in_use[BINARIES_DIR] : "",
	     st->aggregate_b);
  
  if (!files_in_use[which] || strcmp(pn, files_in_use[which])) {
    if (files_in_use[which]) free(files_in_use[which]);
    files_in_use[which] = pn;
    file_changed |= (1 << which);
    gtk_label_set(GTK_LABEL(info_labels[which]), pn);
  }
  
  if (SAMPLES_FILE == which) {
    pn = st->aggregate_b;
    qn = files_in_use[BINARIES_DIR];
    file_to_replace = BINARY_FILE;
    dn = "Open executable file\n%s?";
    asprintf(&dialog_fn,"%s%s", strcmp("/", qn) ? qn : "", pn);
    open_dialog(dn);
    free(dialog_fn);
    maybe_visualise_data();
  }
  else {
    int i;
    GtkTreeIter iter;
    GtkTreeSelection  * select;
    GtkWidget         * suggestion_window  = gtk_scrolled_window_new (NULL, NULL);
    GtkWidget         * suggestion_vbox    = gtk_vbox_new(FALSE, 5);
    GtkWidget         * suggestion_tree    = gtk_tree_view_new ();
    GtkListStore      * suggestion_model   = gtk_list_store_new (1, G_TYPE_STRING);
    GtkCellRenderer   * suggestion_cell    = gtk_cell_renderer_text_new ();
    GtkTreeViewColumn * suggestion_column
      = gtk_tree_view_column_new_with_attributes ("Corresponding Samples",
						  suggestion_cell,
						  "text", 0,
						  NULL);

    hit_list_window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_uposition(hit_list_window, 150, 450 );
    gtk_widget_set_usize (hit_list_window, 400, 150);
    gtk_window_set_title(GTK_WINDOW(hit_list_window), "Matching Samples");
  
    gtk_container_set_border_width(GTK_CONTAINER(suggestion_vbox), 5);
    gtk_container_add(GTK_CONTAINER(hit_list_window), suggestion_vbox);
    gtk_widget_show(suggestion_vbox);

    {
      gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (suggestion_window),
				      GTK_POLICY_AUTOMATIC,
				      GTK_POLICY_AUTOMATIC);

      gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (suggestion_window),
					     suggestion_tree);
      gtk_widget_show (suggestion_tree);
  
      gtk_signal_connect (GTK_OBJECT(suggestion_tree), "row-activated",
			  GTK_SIGNAL_FUNC (suggestion_activated),
			  NULL);
      
      qn = files_in_use[SAMPLES_DIR];
      for (i = 0; i < st->parent->children_next; i++) {
	if (!strcmp(st->parent->children[i]->aggregate_b, st->aggregate_b)) {
	  asprintf(&dn,"%s/%s", strcmp("/", qn) ? qn : "", st->parent->children[i]->aggregate);
	  gtk_list_store_append (GTK_LIST_STORE (suggestion_model), &iter);
	  gtk_list_store_set (GTK_LIST_STORE (suggestion_model),
			      &iter,
			      0, dn,
			      -1);
	  free(dn);
	}
      }

      gtk_tree_view_set_model (GTK_TREE_VIEW (suggestion_tree), GTK_TREE_MODEL (suggestion_model));
      g_object_unref (G_OBJECT (suggestion_model));
    
      select = gtk_tree_view_get_selection(GTK_TREE_VIEW (suggestion_tree));
      gtk_tree_selection_set_mode(select, GTK_SELECTION_SINGLE);
      g_signal_connect(G_OBJECT(select), "changed",
		       G_CALLBACK(suggestion_sel_changed),
		       NULL);

      gtk_tree_view_append_column (GTK_TREE_VIEW (suggestion_tree),
				   GTK_TREE_VIEW_COLUMN (suggestion_column));
  
      gtk_box_pack_start (GTK_BOX (suggestion_vbox), suggestion_window, TRUE, TRUE, 0);
      gtk_widget_show(suggestion_window);
    }

    build_button_box(suggestion_vbox, (long int)SAMPLES_FILE,
		     suggestion_accept_clicked, suggestion_cancel_clicked);
      
    gtk_widget_show(hit_list_window);
  }
}


static samples_tree_s *
get_node_from_path(GtkTreePath * arg1)
{
  int i;
  int depth = gtk_tree_path_get_depth(arg1);
  int * path = gtk_tree_path_get_indices(arg1);
  samples_tree_s * stc = samples_tree_root;
  
  for (i = 1; i < depth; i++) stc = stc->children[path[i]];
  return stc;
}

/*								*/
/* respond to double-click on a sample filename			*/
/*								*/

static void
file_activated (GtkTreeView * treeview __attribute__ ((unused)),
		GtkTreePath * arg1,
		GtkTreeViewColumn * arg2 __attribute__ ((unused)),
		gpointer pwhich)
{
  gint which = (long int)pwhich;
  samples_tree_s * samples_tree_current = get_node_from_path(arg1);
  
  if (0 == samples_tree_current->children_next) {	/* leaf node */
    update_samples_file(which, samples_tree_current);
    gtk_widget_destroy(sample_list_window);
    window_open &= ~(1 << which);
  }
}




static GtkListStore * match_model;

/*								*/
/* traverse and extract aggregate from the samples subtree	*/
/*							*/

static void
extract_samples_tree(samples_tree_s * samples_tree_current)
{
  int i;
  GtkTreeIter iter;

  if (samples_tree_current) {
    if (0 == samples_tree_current->children_next) {
      gtk_list_store_append (GTK_LIST_STORE (match_model), &iter);
      gtk_list_store_set (GTK_LIST_STORE (match_model),
			  &iter,
			  0, samples_tree_current->aggregate,
#ifndef USE_OLD_FASHIONED_SELECT
			  1, samples_tree_current,
#endif
			  -1);
    }
    for (i = 0; i < samples_tree_current->children_next; i++)
      extract_samples_tree(samples_tree_current->children[i]);
  }
}

static gboolean
samples_tree_indexer(GtkTreeModel * model __attribute__ ((unused)),
		     GtkTreePath  * path,
		     GtkTreeIter  * iter  __attribute__ ((unused)),
		     gpointer text)
{
  char * cval;
  samples_tree_s * stc;
  gboolean match_found;
  regex_t preg;
  int rc;
  
  gtk_tree_model_get(model, iter,
		     0, &cval,
		     -1);

#if 1
  if (0 != (rc = regcomp(&preg, (char *)text, REG_NOSUB))) {
#define ERRBUF_SIZE 256
    char * errbuf = alloca(ERRBUF_SIZE);
    regerror(rc,  &preg, errbuf, ERRBUF_SIZE);
    gtk_label_set(GTK_LABEL(status_bar), errbuf);
    return TRUE;
  }
  
  match_found = (0 == regexec(&preg,  cval, 0, NULL, 0)) ? TRUE : FALSE;

#else
  match_found = strcmp(cval, (char *)text) ? FALSE : TRUE;
#endif
  
  if (TRUE == match_found) {
    stc = get_node_from_path(path);
    extract_samples_tree(stc);
  }
  
  return FALSE;
}

static void
match_activated (GtkTreeView * treeview __attribute__ ((unused)),
		 GtkTreePath * arg1 __attribute__ ((unused)),
		 GtkTreeViewColumn * arg2 __attribute__ ((unused)),
		 gpointer pwhich)
{
#ifdef USE_OLD_FASHIONED_SELECT
  if (selected_fn) {
    update_samples_file((long int)pwhich, selected_fn);
    gtk_widget_destroy(sample_list_window);
    gtk_widget_destroy(hit_list_window);
    free(selected_fn);
    selected_fn = 0;
  }
#else
  if (selected_samples_tree) {
    update_samples_file((long int)pwhich, selected_samples_tree);
    gtk_widget_destroy(sample_list_window);
    gtk_widget_destroy(hit_list_window);
    selected_samples_tree = 0;
  }
#endif
}

static void
match_sel_changed(GtkTreeSelection * selection,
		  gpointer data __attribute__ ((unused)))
{
  GtkTreeIter iter;
  GtkTreeModel * model;

  if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
#ifdef USE_OLD_FASHIONED_SELECT
    if (selected_fn) free(selected_fn);
    gtk_tree_model_get(model, &iter, 0, &selected_fn, -1);
#else
    gtk_tree_model_get(model, &iter, 1, &selected_samples_tree, -1);
#endif
  }
}

static void
match_cancel_clicked (GtkObject * object __attribute__ ((unused)),
		       gint       data __attribute__ ((unused)))
{
  gtk_widget_destroy(hit_list_window);
#ifdef USE_OLD_FASHIONED_SELECT
  if (selected_fn) free(selected_fn);
  selected_fn = 0;
#else
  selected_samples_tree = 0;
#endif
}

static void
match_accept_clicked (GtkButton * button __attribute__ ((unused)),
		       gpointer data __attribute__ ((unused)))
{
  match_activated(NULL, NULL, NULL, (gpointer)global_which);
}

static void
search_clicked(GtkButton * button __attribute__ ((unused)),
	       gpointer data)
{
  GtkTreeModel * file_model = gtk_tree_view_get_model(GTK_TREE_VIEW(files_tree));
  const char * text = gtk_entry_get_text(GTK_ENTRY(data));
  GtkWidget * vbox = gtk_vbox_new(FALSE, 5);
  GtkWidget * scrolled_window  = gtk_scrolled_window_new (NULL, NULL);
  GtkWidget * match_tree = gtk_tree_view_new ();
  GtkCellRenderer * cell = gtk_cell_renderer_text_new ();
  GtkTreeViewColumn * column = gtk_tree_view_column_new_with_attributes ("Matches",
									 cell,
									 "text", 0,
									 NULL);
  GtkTreeSelection * select;
  
  hit_list_window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_uposition(hit_list_window, 150, 450 );
  gtk_widget_set_usize (hit_list_window, 400, 150);
  gtk_window_set_title(GTK_WINDOW(hit_list_window), "Matching Files");
  
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_container_add(GTK_CONTAINER(hit_list_window), vbox);
  gtk_widget_show(vbox);

  {
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
				    GTK_POLICY_AUTOMATIC,
				    GTK_POLICY_AUTOMATIC);

    gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrolled_window),
					   match_tree);
    gtk_widget_show (match_tree);
  
#ifdef USE_OLD_FASHIONED_SELECT
    match_model = gtk_list_store_new (1, G_TYPE_STRING);
#else
    match_model = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_INT);
#endif
  
    gtk_signal_connect (GTK_OBJECT(match_tree), "row-activated",
			GTK_SIGNAL_FUNC (match_activated),
			(gpointer)global_which);

    gtk_tree_model_foreach(GTK_TREE_MODEL (file_model),samples_tree_indexer, (gpointer)text);
    gtk_tree_view_set_model (GTK_TREE_VIEW (match_tree), GTK_TREE_MODEL (match_model));
    g_object_unref (G_OBJECT (match_model));
    
    select = gtk_tree_view_get_selection(GTK_TREE_VIEW (match_tree));
    gtk_tree_selection_set_mode(select, GTK_SELECTION_SINGLE);
    g_signal_connect(G_OBJECT(select), "changed",
		     G_CALLBACK(match_sel_changed),
		     NULL);

    gtk_tree_view_append_column (GTK_TREE_VIEW (match_tree),
				 GTK_TREE_VIEW_COLUMN (column));
  
    gtk_box_pack_start (GTK_BOX (vbox), scrolled_window, TRUE, TRUE, 0);
    gtk_widget_show(scrolled_window);
  }

  build_button_box(vbox, global_which, match_accept_clicked, match_cancel_clicked);
      
  gtk_widget_show(hit_list_window);
}

static void
sample_cancel_clicked (GtkObject * object __attribute__ ((unused)),
		       gpointer        pwhich)
{
  free_samples_tree(samples_tree_root);
  samples_tree_root = NULL;
  gtk_widget_destroy(sample_list_window);
  window_open &= ~(1 << (long int)pwhich);
}

samples_tree_s * samples_tree_selected = 0;

static void
sample_accept_clicked (GtkButton * button __attribute__ ((unused)),
		       gpointer pwhich)
{
  if (samples_tree_selected && (0 == samples_tree_selected->children_next)) {	/* leaf node */
    update_samples_file((long int)pwhich, samples_tree_selected);
    gtk_widget_destroy(sample_list_window);
    window_open &= ~(1 << (long int)pwhich);
#if 0
#ifndef DONT_USE_OBSTACKS
    fprintf(stderr, "obstack_free in sample_accept_clicked\n");
    obstack_free(&history_obstack, NULL);
    obstack_free(&samples_tree_obstack, NULL);
    obstack_free(&samples_tree_element_obstack, NULL);
    obstack_free(&dir_obstack, NULL);
    obstack_free(&dir_ptr_obstack, NULL);
#endif
#endif
  }
}

static void
file_sel_changed(GtkTreeSelection * selection,
		  gpointer data __attribute__ ((unused)))
{
  GtkTreeIter iter;
  GtkTreeModel * model;
  GtkTreePath * path;

  if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
    path = gtk_tree_model_get_path(model, &iter);
    samples_tree_selected = get_node_from_path(path);
  }
}


/*									*/
/*									*/
/*									*/
/* sample select main							*/
/*									*/
/*									*/
/*									*/

void
handle_open_sample(GtkWidget * w __attribute__ ((unused)),
		   gint which __attribute__ ((unused)),
		   void * callback_action __attribute__ ((unused)))
{
  GtkWidget * vbox;
  
  if (!(window_open & (1 << which))) {
    window_open |= (1 << which);

    global_which = (long int)which;
    switch(which) {
    case SAMPLES_FILE:	which_dir = SAMPLES_DIR;	break;
    case BINARY_FILE:	which_dir = BINARIES_DIR;	break;
    }

    sample_list_window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
    gtk_signal_connect(GTK_OBJECT(sample_list_window),
                     "destroy",
                     GTK_SIGNAL_FUNC(sample_list_cancel),
                     (gpointer)global_which);
    gtk_widget_set_uposition(sample_list_window, 100, 400 );
    gtk_widget_set_usize (sample_list_window, 300, 500);
    gtk_window_set_title(GTK_WINDOW(sample_list_window),
			 (SAMPLES_FILE == which) ? "Samples Files" : "Binary Files");

    vbox = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
    gtk_container_add(GTK_CONTAINER(sample_list_window), vbox);
    gtk_widget_show(vbox);
    
    {				/* create history selector	*/
      history_options_menu = gtk_option_menu_new ();

      build_history_table(which);

      populate_history_menu();
      gtk_box_pack_start (GTK_BOX (vbox), history_options_menu, FALSE, FALSE, 0);
      gtk_widget_show (history_options_menu);
    }

    {				/* search */
      GtkWidget * frame  = gtk_frame_new("Find files (regex)");
      GtkWidget * shbox  = gtk_hbox_new(FALSE, 2);
      GtkWidget * button = gtk_button_new_with_label("Search");
      GtkWidget * entry  = gtk_entry_new();
      
      
      gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 0);
      
      gtk_widget_show(shbox);

      gtk_signal_connect (GTK_OBJECT(button), "clicked",
			  GTK_SIGNAL_FUNC (search_clicked),
			  (gpointer)entry);

      gtk_signal_connect (GTK_OBJECT(entry), "activate",
			  GTK_SIGNAL_FUNC (search_clicked),
			  (gpointer)entry);

      gtk_box_pack_start (GTK_BOX (shbox), entry, TRUE, TRUE, 0);
      gtk_widget_show (entry);

      gtk_widget_grab_focus(entry);
      
      gtk_box_pack_start (GTK_BOX (shbox), button, FALSE, FALSE, 0);
      gtk_widget_show (button);

      gtk_container_add (GTK_CONTAINER (frame), shbox);
      gtk_frame_set_label_align (GTK_FRAME (frame), 0.0, 0.0);
      gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);
 
      gtk_widget_show(frame);
    }
    
    {				/* dir/file hpane */
      GtkWidget * hpane = gtk_hpaned_new();
      GtkWidget * dir_list;
      GtkTreeViewColumn * column;
      GtkCellRenderer * renderer;
      GtkWidget * tree_window_scroll;
      GtkTreeSelection * select;

      build_samples_tree(files_in_use[SAMPLES_DIR]  ? : DEFAULT_SAMPLES_DIR,
			 (BINARIES_DIR == which_dir)
			 ? (files_in_use[BINARIES_DIR] ? : DEFAULT_BINARIES_DIR)
			 : NULL);

      tree_window_scroll = gtk_scrolled_window_new (NULL, NULL);
      gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (tree_window_scroll),
				      GTK_POLICY_AUTOMATIC,
				      GTK_POLICY_AUTOMATIC);

      files_tree = gtk_tree_view_new();
      select = gtk_tree_view_get_selection(GTK_TREE_VIEW (files_tree));
      gtk_tree_selection_set_mode(select, GTK_SELECTION_SINGLE);
      g_signal_connect(G_OBJECT(select), "changed",
		       G_CALLBACK(file_sel_changed),
		       NULL);

      gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (tree_window_scroll),
					     files_tree);
      gtk_widget_show (files_tree);
      gtk_signal_connect (GTK_OBJECT(files_tree), "row-activated",
			  GTK_SIGNAL_FUNC (file_activated),
			  (gpointer)global_which);
  
      renderer = gtk_cell_renderer_text_new ();

      column = gtk_tree_view_column_new_with_attributes ("File", renderer,
							 "text", 0,
							 NULL);

      gtk_tree_view_append_column (GTK_TREE_VIEW (files_tree), column);

      set_files_tree_list();

      gtk_tree_view_expand_all(GTK_TREE_VIEW(files_tree));
      
      dir_list  = create_subdirs_list (dir_count, dir_list_ptrs);
      
      gtk_box_pack_start (GTK_BOX (vbox), hpane, TRUE, TRUE, 0);
      gtk_widget_show(hpane);
      
      gtk_paned_add1 (GTK_PANED (hpane), dir_list);
      gtk_widget_show (dir_list);

      gtk_paned_add2 (GTK_PANED (hpane), tree_window_scroll);
      gtk_widget_show (tree_window_scroll);
    }

    build_button_box(vbox, (long int)which, sample_accept_clicked, sample_cancel_clicked);
      
    gtk_widget_show(sample_list_window);

  }
}


