/******************************************************************************
Copyright (C) 2002, 2003 Red Hat, Inc.

This file is part of op_visualise, a part of Oprofile.

This file may be redistributed and/or modifed under the terms of the GNU
General Public License as published by the Free Software Foundation; either
version 2, or (at your option) any later version.

op_visualise is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details. 

op_visualise was written, and is maintained, by Chris Moller,
Red Hat Inc, cmoller@redhat.com

This file provides functions that analyse and display oprofile data.

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#define __USE_ISOC99 1
#include <math.h>

#include "op_visualise.h"

#include "elf/common.h"
#include "elf/internal.h"

/* #define USE_FUNCTION_TREE_SHOW */

extern GtkWidget * status_bar;

static gint base_y;

#ifndef DONT_USE_OBSTACKS
static struct obstack functions_obstack;
static int            functions_obstack_initted = 0;
static struct obstack symbols_obstack;
static struct obstack symbols_index_obstack;
static int            symbols_obstack_initted = 0;
static struct obstack text_section_obstack;
static struct obstack text_section_index_obstack;
static int            text_section_obstack_initted = 0;
#endif

int
function_tree_compare(gconstpointer a, gconstpointer b)
{
  return strcmp((char *)a, (char *)b);
}

#ifdef USE_FUNCTION_TREE_SHOW
static gboolean
function_tree_show(gpointer key, gpointer value, gpointer data __attribute__ ((unused)))
{
  fprintf(stderr,"tree key = \"%s\"\n", (char *)key);
  fprintf(stderr,"tree value = \"%s\", %#010lx, %#010lx\n",
	  ((function_data_s *)value)->function_name,
	  ((function_data_s *)value)->min_loc,
	  ((function_data_s *)value)->max_loc);
  return FALSE;
}
#endif

static void
line_callback(bfd_vma lpos,
	      unsigned int value)
{
  int i;
  bfd_vma spos;
  bfd_vma ipos;
  const char *filename, *functionname;
  unsigned int lineno;
  int sfn;
  function_data_s * hit;
  function_data_s * fdt;

  for ( i = 0; i < text_section_next; i++) {
      if ((lpos >= (spos = bfd_get_section_vma(ibfd, text_section_list[i]))) &&
	  (lpos < (spos + bfd_section_size(ibfd, text_section_list[i])))) {
	if (bfd_find_nearest_line(ibfd, text_section_list[i], syms,
				  ipos = lpos - bfd_get_section_vma(ibfd, text_section_list[i]),
				  &filename, &functionname, &lineno)) {
	  /* fixme -- add test for .name */
	  if (functionname && (0 < (sfn = strlen(functionname)))
	      && ('.' != *functionname)
	      ) {

	    if ((hit = g_tree_lookup(function_tree, functionname))) {
	      if (lpos < hit->min_loc) hit->min_loc = lpos;
	      else if (lpos > hit->max_loc) hit->max_loc = lpos;
	      hit->total_hits += value;
	    }
	    else {
#ifndef DONT_USE_OBSTACKS
	      fdt = obstack_alloc(&functions_obstack, sizeof(function_data_s));
	      fdt->function_name = obstack_copy0(&functions_obstack,
						 functionname, sfn);
#endif
	      fdt->min_loc = fdt->max_loc = lpos;
	      fdt->total_hits = value;
	      g_tree_insert(function_tree, (gpointer)functionname, fdt);
	    }
	  }
	  if (filename && (0 < (sfn = strlen(filename)))) {
	    if ((hit = g_tree_lookup(file_tree, filename))) {
	      if (lpos < hit->min_loc) hit->min_loc = lpos;
	      else if (lpos > hit->max_loc) hit->max_loc = lpos;
	      hit->total_hits += value;
	    }
	    else {
#ifndef DONT_USE_OBSTACKS
	      fdt = obstack_alloc(&functions_obstack, sizeof(function_data_s));
	      fdt->function_name = obstack_copy0(&functions_obstack,
						 filename, sfn);
#endif
	      fdt->min_loc = fdt->max_loc = lpos;
	      fdt->total_hits = value;
	      g_tree_insert(file_tree, (gpointer)filename, fdt);
	    }
	  }
	}
      }
  }
}

void
display_callback(bfd_vma lpos,
		 unsigned int value,
		 GdkGC * gc)
{
  gint x = (gint)lrint( x_scale * ((double)lpos - min_ruler_d));
  gint y = (gint)lrint( y_scale * ((double)value));

  gdk_draw_line(pixmap, gc, x, base_y, x, base_y - y);
}

static int
symbol_list_sort_fcn( const void * a, const void * b)
{
  symvalue aa = (*(symbol_info **)a)->value;
  symvalue bb = (*(symbol_info **)b)->value;
  return (aa < bb) ? -1 : ((aa > bb) ? 1 : 0);
}

static int
text_section_list_sort_fcn( const void * a, const void * b)
{
  bfd_size_type a_size = bfd_section_size(ibfd, (*(asection **)a));
  bfd_size_type b_size = bfd_section_size(ibfd, (*(asection **)b));
    
  return (a_size > b_size) ? -1 : ((a_size < b_size) ? 1 : 0);
}

static void
lookat_section(
	       bfd * abfd __attribute__ ((unused)),
	       asection * asect,
	       PTR optarg __attribute__ ((unused)))
{
  void * osp;
  
  if (bfd_get_section_flags (ibfd, asect) & SEC_CODE) {
#ifndef DONT_USE_OBSTACKS
    osp = obstack_copy (&text_section_obstack, asect, sizeof(asection));
    obstack_ptr_grow(&text_section_index_obstack, osp);
    text_section_next++;
#endif
  }
}

static int
open_binary_file()
{
  char ** matching;
  long phdr_upper_bound;
  long phdr_count;
  Elf_Internal_Phdr * phdrs;
  int ld_hdr;
  
  if (ibfd) bfd_close(ibfd);

  file_changed &= ~(1 << BINARY_FILE);
    
  if (NULL == (ibfd = bfd_openr(files_in_use[BINARY_FILE], NULL)))
    gtk_label_set(GTK_LABEL(status_bar), "Error opening binary file.");
  else if (!bfd_check_format_matches(ibfd, bfd_object, &matching)) {
    gtk_label_set(GTK_LABEL(status_bar), "Format failure of binary file.");
    ibfd =  NULL;
  }

  if (NULL == ibfd) return 0;
    
#ifndef DONT_USE_OBSTACKS
  if (text_section_obstack_initted) {
    obstack_free(&text_section_obstack, NULL);
    obstack_free(&text_section_index_obstack, NULL);
  }
  obstack_init(&text_section_obstack);
  obstack_init(&text_section_index_obstack);
  text_section_obstack_initted = 1;
#endif

  text_section_next = 0;
  bfd_map_over_sections (ibfd, lookat_section, NULL);
      
#ifndef DONT_USE_OBSTACKS
  text_section_list = obstack_finish(&text_section_index_obstack);
#endif

  /* sort for biggest section first to increase odds of first-pass hit */
  qsort(text_section_list, text_section_next, sizeof(asection *), text_section_list_sort_fcn);

  phdr_upper_bound =  bfd_get_elf_phdr_upper_bound(ibfd);

  phdrs = xmalloc(phdr_upper_bound);
  phdr_count = bfd_get_elf_phdrs(ibfd, phdrs);

  for (ld_hdr = 0; ld_hdr < phdr_count; ld_hdr++) {
    if ((phdrs[ld_hdr].p_type == PT_LOAD) &&
	(phdrs[ld_hdr].p_flags == (PF_R | PF_X))) {
      start_vma = phdrs[ld_hdr].p_paddr;
      break;
    }
  }

  {
    long symcount;
    PTR minisyms;
    unsigned int size;
    bfd_byte *from, *fromend, *to;
    asymbol *store;
    symbol_info syminfo;
    static bfd *lineno_cache_bfd = NULL;
    
#ifndef DONT_USE_OBSTACKS
    if (symbols_obstack_initted) {
      obstack_free(&symbols_obstack, NULL);
      obstack_free(&symbols_index_obstack, NULL);
    }
    obstack_init(&symbols_obstack);
    obstack_init(&symbols_index_obstack);
    symbols_obstack_initted = 1;
    
    if (functions_obstack_initted) obstack_free(&functions_obstack, NULL);
    obstack_init(&functions_obstack);
    functions_obstack_initted = 1;
#endif

    symcount = bfd_read_minisymbols (ibfd, 0, &minisyms, &size);

    store = bfd_make_empty_symbol (ibfd);
    
    from = (bfd_byte *) minisyms;
    fromend = from + symcount * size;
    to = (bfd_byte *) minisyms;

    sym_count = 0;

    if (ibfd != lineno_cache_bfd && syms != NULL) {
      free (syms);
      syms = NULL;
    }
    if (syms == NULL) {
      long symsize;
      
      symsize = bfd_get_symtab_upper_bound (ibfd);
      if (symsize < 0) {
	fprintf(stderr, "oops 1\n");	/* fixme */
      }
      else {
	syms = (asymbol **) xmalloc (symsize);
	symcount = bfd_canonicalize_symtab (ibfd, syms);
	if (symcount < 0) {
	  fprintf(stderr, "oops 2\n");	/* fixme */
	}
	lineno_cache_bfd = ibfd;
      }
    }

    if (syms) {
      double total_progress;
      double tick_progress;
      double tick_threshold;

      if (symcount > PROGRESS_BAR_THRESHOLD) {
	initialise_progress_bar((double)symcount);
	long_op_in_progress = TRUE;
      }
      
      for (total_progress = 0.0, tick_progress = 0.0,
	     tick_threshold = PROGRESS_BAR_TICK * (double)symcount;
	   from < fromend;
	   from += size, total_progress += 1.0, tick_progress += 1.0) {
	asymbol *sym;
	void * osp;

	sym = bfd_minisymbol_to_symbol (ibfd, 0, (const PTR) from, store);	

	if ((0 != sym->value)
#if 1
	    && (bfd_get_section_flags (ibfd, bfd_get_section (sym)) & SEC_CODE)
#endif
	    ) {
	    
	  const char *filename, *functionname;
	  unsigned int lineno;
	  asection * asec;
	    
	  filename = 0;
	  functionname = 0;
	  lineno = 0;
	  asec = bfd_get_section (sym);

	  /* fixme -- i have no idea why, but if this statement is omitted the text */
	  /* asect name fields (text_section_list) get nulled out. */
	  bfd_find_nearest_line(ibfd, asec, syms, sym->value,
				&filename, &functionname, &lineno);
#if 0
	  fprintf(stderr,"\bsym:\n");
	  fprintf(stderr,"\tsymbol name = %s\n", bfd_asymbol_name(sym));
	  fprintf(stderr,"\tsymbol value = %#010lx\n", bfd_asymbol_value(sym));
	  fprintf(stderr,"\tsection name = %s\n", bfd_get_section_name(ibfd, bfd_get_section (sym)));
	  fprintf(stderr,"\tfilename = %s\n", filename);
	  fprintf(stderr,"\tfunctionname = %s\n", functionname);
	  fprintf(stderr,"\tlineno = %d\n\n", lineno);
#endif	

	  bfd_get_symbol_info (ibfd, sym, &syminfo);
	  if (0 != syminfo.value) {
#if 0
	    fprintf(stderr,"syminfo:\n");
	    fprintf(stderr,"\t%08lx %c %s\n",
		    syminfo.value,
		    (int)(syminfo.type),
		    syminfo.name);
#endif	
#ifndef DONT_USE_OBSTACKS
	    osp = obstack_copy (&symbols_obstack, &syminfo, sizeof(syminfo));
	    obstack_ptr_grow(&symbols_index_obstack, osp);
#endif
	    sym_count++;
	  }


	}
        if (symcount > PROGRESS_BAR_THRESHOLD) {
	  if (tick_progress >= tick_threshold) {
            tick_progress = 0.0;
	    if (update_progress_bar(total_progress)) break;;
	  }
	}
      }
      if (symcount > PROGRESS_BAR_THRESHOLD) {
	remove_progress_bar();
	long_op_in_progress = FALSE;
      }
    }
      
#ifndef DONT_USE_OBSTACKS
    symbol_list = obstack_finish(&symbols_index_obstack);
#endif
      
    qsort(symbol_list, sym_count, sizeof(symbol_info *), symbol_list_sort_fcn);
  }

#if 0
  {
    int i;
      
    fprintf(stderr,"sym_count = %d\n",sym_count);
    for ( i = 0; i < sym_count; i++) {
      fprintf(stderr,"%08lx: [%6lu] %c %s\n",
	      symbol_list[i]->value,
	      (i < (sym_count - 1)) ? (symbol_list[i+1]->value - symbol_list[i]->value) : 0xffffffff,
	      (int)(symbol_list[i]->type),
	      symbol_list[i]->name);
    }
  }
#endif

  return 1;
}

static int
open_samples_file()
{
  struct stat buf;
  int csf_rc;
  int rc = 1;

  if (0 != stat(files_in_use[SAMPLES_FILE], &buf)) {
    gtk_label_set(GTK_LABEL(status_bar), "Error stating samples file.");
    rc = 0;
  }

  if (rc && !S_ISREG(buf.st_mode)) {
    gtk_label_set(GTK_LABEL(status_bar), "Non-regular file selected for samples.");
    rc = 0;
  }
  
  if (rc && !(S_IROTH & buf.st_mode)) {
    gtk_label_set(GTK_LABEL(status_bar), "Samples file not readable.");
    rc = 0;
  }

  if (!rc) return 0;

  csf_rc = copy_sample_file(files_in_use[SAMPLES_FILE]);
  if (EXIT_SUCCESS != csf_rc) {
    gtk_label_set(GTK_LABEL(status_bar), "Error opening samples file.");
    return 0;
  }

  max_value_d = (double)max_value;
  max_value_l = max_value;
  GTK_ADJUSTMENT(adjustment)->value =
    GTK_ADJUSTMENT(adjustment)->lower = min_key_d = min_ruler_d = (double)min_key;
  GTK_ADJUSTMENT(adjustment)->upper = max_key_d = max_ruler_d = (double)max_key;
  GTK_ADJUSTMENT(adjustment)->page_size = max_key_d - min_key_d;
  gtk_adjustment_changed(GTK_ADJUSTMENT(adjustment));
  update_labels();
  
  x_scale = ((double)drawing_area->allocation.width)/(max_key_d - min_key_d);
  y_scale = ((double)drawing_area->allocation.height)/max_value_d;
  
  gtk_ruler_set_range(GTK_RULER(vrule), max_value_d, 0.0, max_value_d, 0.0);

  base_y = drawing_area->allocation.height;

  return 1;
}

void
draw_chart()
{
  long int min_l, max_l;
  
  gdk_draw_rectangle(pixmap, drawing_area->style->black_gc,
		     TRUE,
		     0, 0,
		     drawing_area->allocation.width,
		     drawing_area->allocation.height);
  if (highlight_active) {
    gdk_draw_rectangle(pixmap, yellow_gc,
  		       TRUE,
  		       (gint)lrint( x_scale * ((double)highlight_min - min_ruler_d)),
		       0,
  		       1 + (gint)lrint( x_scale * (double)(highlight_max - highlight_min)),
		       drawing_area->allocation.height);
  }

  min_l = lrint(min_ruler_d) - 1;
  max_l = lrint(max_ruler_d) + 1;
  
  samples_travel(min_l, max_l, display_callback, drawing_area->style->white_gc);

  if (symbol_list) {
    int i;
    gint x;
      
    for ( i = 0; i < sym_count; i++) {
      x = (gint)lrint( x_scale * ((double)symbol_list[i]->value - min_ruler_d));
      gdk_draw_line(pixmap, red_gc, x,0, x, 10);
    }
  }
  

#ifdef USE_ANNOTATION
  gdk_draw_string (pixmap,					/* GdkDrawable *drawable 	*/
		   font,					/* GdkFont     *font		*/
		   drawing_area->style->white_gc,		/* GdkGC       *gc		*/
		   300,						/* gint         x		*/
		   300,						/* gint         y		*/
		   "hi there");					/* const gchar *string		*/
#endif

  gtk_widget_draw (drawing_area, NULL);
}

void
maybe_visualise_data()
{
  int rc = 1;

  if (files_in_use[SAMPLES_FILE]
      && files_in_use[BINARY_FILE]
      /* && files_in_use[SOURCES_DIR] */ ) {
    
    if (file_changed & (1 << BINARY_FILE)) {
      rc &= open_binary_file();
#if 0
      fprintf(stderr, "rc = %d for binary open of %s\n", rc, files_in_use[BINARY_FILE]);
#endif
    }
    if (file_changed & (1 << SAMPLES_FILE)) {
      rc &= open_samples_file();
#if 0
      fprintf(stderr, "rc = %d for samples open of %s\n", rc, files_in_use[SAMPLES_FILE]);
#endif
    }

    if (function_tree) g_tree_destroy(function_tree);
    function_tree = g_tree_new(function_tree_compare);
    if (file_tree) g_tree_destroy(file_tree);
    file_tree = g_tree_new(function_tree_compare);

    if (rc) {
      samples_travel(0, ~0, line_callback, drawing_area->style->white_gc);
#ifdef USE_FUNCTION_TREE_SHOW
      fprintf(stderr,"function tree nnodes = %d\n", g_tree_nnodes(function_tree));
      g_tree_foreach(function_tree, function_tree_show, NULL);
      fprintf(stderr,"file tree nnodes = %d\n", g_tree_nnodes(file_tree));
      g_tree_foreach(file_tree, function_tree_show, NULL);
#endif
      draw_chart();
    }
    else {
      gdk_draw_rectangle(pixmap, drawing_area->style->black_gc,
			 TRUE,
			 0, 0,
			 drawing_area->allocation.width,
			 drawing_area->allocation.height);
      if (highlight_active) {
        gdk_draw_rectangle(pixmap, yellow_gc,
  		           TRUE,
  		           (gint)lrint( x_scale * ((double)highlight_min - min_ruler_d)),
		           0,
  		           1 + (gint)lrint( x_scale * (double)(highlight_max - highlight_min)),
		           drawing_area->allocation.height);
      }
      gtk_widget_draw (drawing_area, NULL);
    }
  }
}
